import os
import io
import sys
import time
import shlex
import atexit
import threading
import traceback
import subprocess
import contextlib

import RNS
import LXMF
import nomadnet

from nomadnet.Directory import DirectoryEntry
from datetime import datetime

import RNS.vendor.umsgpack as msgpack

from ._version import __version__
from RNS.vendor.configobj import ConfigObj

class NomadNetworkApp:
    time_format      = "%Y-%m-%d %H:%M:%S"
    _shared_instance = None

    userdir = os.path.expanduser("~")
    if os.path.isdir("/etc/nomadnetwork") and os.path.isfile("/etc/nomadnetwork/config"):
        configdir = "/etc/nomadnetwork"
    elif os.path.isdir(userdir+"/.config/nomadnetwork") and os.path.isfile(userdir+"/.config/nomadnetwork/config"):
        configdir = userdir+"/.config/nomadnetwork"
    else:
        configdir = userdir+"/.nomadnetwork"

    START_ANNOUNCE_DELAY = 3

    def exit_handler(self):
        self.should_run_jobs = False

        RNS.log("Saving directory...", RNS.LOG_VERBOSE)
        self.directory.save_to_disk()

        if hasattr(self, "rrc") and self.rrc is not None:
            try:
                self.rrc.save()
                self.rrc.shutdown()
            except Exception:
                pass

        if hasattr(self.ui, "restore_ixon"):
            if self.ui.restore_ixon:
                try:
                    os.system("stty ixon")

                except Exception as e:
                    RNS.log("Could not restore flow control sequences. The contained exception was: "+str(e), RNS.LOG_WARNING)

        if hasattr(self.ui, "restore_palette"):
            if self.ui.restore_palette:
                try:
                    self.ui.screen.write("\x1b]104\x07")

                except Exception as e:
                    RNS.log("Could not restore terminal color palette. The contained exception was: "+str(e), RNS.LOG_WARNING)

        RNS.log("Nomad Network Client exiting now", RNS.LOG_VERBOSE)

    def exception_handler(self, e_type, e_value, e_traceback):
        RNS.log("An unhandled exception occurred, the details of which will be dumped below", RNS.LOG_ERROR)
        RNS.log("Type  : "+str(e_type), RNS.LOG_ERROR)
        RNS.log("Value : "+str(e_value), RNS.LOG_ERROR)
        t_string = ""
        for line in traceback.format_tb(e_traceback):
            t_string += line
        RNS.log("Trace : \n"+t_string, RNS.LOG_ERROR)

        if issubclass(e_type, KeyboardInterrupt):
            sys.__excepthook__(e_type, e_value, e_traceback)

    def __init__(self, configdir = None, rnsconfigdir = None, daemon = False, force_console = False):
        self.version       = __version__
        self.enable_client = False
        self.enable_node   = False
        self.identity      = None

        self.uimode        = None

        self.announce_interval = 6*60*60

        if configdir == None:
            self.configdir = NomadNetworkApp.configdir
        else:
            self.configdir = configdir

        if force_console:
            self.force_console_log = True
        else:
            self.force_console_log = False

        if NomadNetworkApp._shared_instance == None:
            NomadNetworkApp._shared_instance = self

        self.rns = RNS.Reticulum(configdir = rnsconfigdir)

        self.configpath        = self.configdir+"/config"
        self.ignoredpath       = self.configdir+"/ignored"
        self.logfilepath       = self.configdir+"/logfile"
        self.errorfilepath     = self.configdir+"/errors"
        self.pnannouncedpath   = self.configdir+"/pnannounced"
        self.storagepath       = self.configdir+"/storage"
        self.identitypath      = self.configdir+"/storage/identity"
        self.cachepath         = self.configdir+"/storage/cache"
        self.resourcepath      = self.configdir+"/storage/resources"
        self.conversationpath  = self.configdir+"/storage/conversations"
        self.directorypath     = self.configdir+"/storage/directory"
        self.peersettingspath  = self.configdir+"/storage/peersettings"
        self.tmpfilespath      = self.configdir+"/storage/tmp"
        self.attachmentpath    = self.configdir+"/storage/attachments"

        self.pagespath         = self.configdir+"/storage/pages"
        self.filespath         = self.configdir+"/storage/files"
        self.cachepath         = self.configdir+"/storage/cache"
        self.examplespath      = self.configdir+"/examples"

        self.downloads_path    = os.path.expanduser("~/Downloads")
        self.attachment_save_path = None

        self.firstrun               = False
        self.should_run_jobs        = True
        self.job_interval           = 5
        self.defer_jobs             = 90
        self.page_refresh_interval  = 0
        self.file_refresh_interval  = 0

        self.static_peers            = []
        self.peer_announce_at_start  = True
        self.try_propagation_on_fail = True
        self.disable_propagation     = True
        self.notify_on_new_message   = True
        self.compose_markdown        = True

        self.lxmf_max_propagation_size = None
        self.lxmf_max_sync_size        = None
        self.lxmf_max_incoming_size    = None
        self.node_propagation_cost     = LXMF.LXMRouter.PROPAGATION_COST

        self.periodic_lxmf_sync = True
        self.lxmf_sync_interval = 360*60
        self.lxmf_sync_limit    = 8
        self.compact_stream     = False

        self.required_stamp_cost   = None
        self.accept_invalid_stamps = False

        self.rrc_history_per_room_cap = 500
        self.rrc_filter_loaded_history = True
        self.rrc_ephemeral_notices = 600
        self.rrc_nick_colors = True
        self.rrc_nick_colors_theme = None
        self.rrc_mention_color = None
        self.rrc_color_mention_timestamps = True
        self.rrc_ui_justify_msgs = True
        self.rrc_ui_space_msgs = False
        self.rrc_ui_render_markdown = True
        self.rrc_ui_render_micron = True
        self.rrc_show_gutters = False
        self.rrc_enable_esoterics = False

        if not os.path.isdir(self.storagepath):
            os.makedirs(self.storagepath)

        if not os.path.isdir(self.cachepath):
            os.makedirs(self.cachepath)

        if not os.path.isdir(self.resourcepath):
            os.makedirs(self.resourcepath)

        if not os.path.isdir(self.conversationpath):
            os.makedirs(self.conversationpath)

        if not os.path.isdir(self.pagespath):
            os.makedirs(self.pagespath)

        if not os.path.isdir(self.filespath):
            os.makedirs(self.filespath)

        if not os.path.isdir(self.cachepath):
            os.makedirs(self.cachepath)

        if not os.path.isdir(self.tmpfilespath):
            os.makedirs(self.tmpfilespath)

        if not os.path.isdir(self.attachmentpath):
            os.makedirs(self.attachmentpath)
        else:
            self.clear_tmp_dir()

        if os.path.isfile(self.configpath):
            try:
                self.config = ConfigObj(self.configpath)
                try:
                    self.applyConfig()
                except Exception as e:
                    RNS.log("The configuration file is invalid. The contained exception was: "+str(e), RNS.LOG_ERROR)
                    nomadnet.panic()

                RNS.log("Configuration loaded from "+self.configpath)
            except Exception as e:
                RNS.log("Could not parse the configuration at "+self.configpath, RNS.LOG_ERROR)
                RNS.log("Check your configuration file for errors!", RNS.LOG_ERROR)
                nomadnet.panic()
        else:
            if not os.path.isdir(self.examplespath):
                try:
                    import shutil
                    examplespath = os.path.join(os.path.dirname(__file__), "examples")
                    shutil.copytree(examplespath, self.examplespath, ignore=shutil.ignore_patterns("__pycache__"))

                except Exception as e:
                    RNS.log("Could not copy examples into the "+self.examplespath+" directory.", RNS.LOG_ERROR)
                    RNS.log("The contained exception was: "+str(e), RNS.LOG_ERROR)

            RNS.log("Could not load config file, creating default configuration file...")
            self.createDefaultConfig()
            self.firstrun = True

        if os.path.isfile(self.identitypath):
            try:
                self.identity = RNS.Identity.from_file(self.identitypath)
                if self.identity != None:
                    RNS.log("Loaded Primary Identity %s from %s" % (str(self.identity), self.identitypath))
                else:
                    RNS.log("Could not load the Primary Identity from "+self.identitypath, RNS.LOG_ERROR)
                    nomadnet.panic()
            except Exception as e:
                RNS.log("Could not load the Primary Identity from "+self.identitypath, RNS.LOG_ERROR)
                RNS.log("The contained exception was: %s" % (str(e)), RNS.LOG_ERROR)
                nomadnet.panic()
        else:
            try:
                RNS.log("No Primary Identity file found, creating new...")
                self.identity = RNS.Identity()
                self.identity.to_file(self.identitypath)
                RNS.log("Created new Primary Identity %s" % (str(self.identity)))
            except Exception as e:
                RNS.log("Could not create and save a new Primary Identity", RNS.LOG_ERROR)
                RNS.log("The contained exception was: %s" % (str(e)), RNS.LOG_ERROR)
                nomadnet.panic()

        if os.path.isfile(self.peersettingspath):
            try:
                file = open(self.peersettingspath, "rb")
                self.peer_settings = msgpack.unpackb(file.read())
                file.close()

                if not "node_last_announce" in self.peer_settings:
                    self.peer_settings["node_last_announce"] = None

                if not "propagation_node" in self.peer_settings:
                    self.peer_settings["propagation_node"] = None

                if not "last_lxmf_sync" in self.peer_settings:
                    self.peer_settings["last_lxmf_sync"] = 0

                if not "node_connects" in self.peer_settings:
                    self.peer_settings["node_connects"] = 0

                if not "served_page_requests" in self.peer_settings:
                    self.peer_settings["served_page_requests"] = 0

                if not "served_file_requests" in self.peer_settings:
                    self.peer_settings["served_file_requests"] = 0

                self.peer_settings["announce_interval"] = self.announce_interval

            except Exception as e:
                RNS.logdest = RNS.LOG_STDOUT
                RNS.log(f"Could not load local peer settings from {self.peersettingspath}", RNS.LOG_ERROR)
                RNS.log(f"The contained exception was: {e}", RNS.LOG_ERROR)
                RNS.log(f"This likely means that the peer settings file has become corrupt.", RNS.LOG_ERROR)
                RNS.log(f"You can try deleting the file at {self.peersettingspath} and restarting nomadnet.", RNS.LOG_ERROR)
                nomadnet.panic()
        else:
            try:
                RNS.log("No peer settings file found, creating new...")
                self.peer_settings = {
                    "display_name": "Anonymous Peer",
                    "announce_interval": self.announce_interval,
                    "last_announce": None,
                    "node_last_announce": None,
                    "propagation_node": None,
                    "last_lxmf_sync": 0,
                    "node_connects": 0,
                    "served_page_requests": 0,
                    "served_file_requests": 0
                }
                self.save_peer_settings()
                RNS.log("Created new peer settings file")
            except Exception as e:
                RNS.log("Could not create and save a new peer settings file", RNS.LOG_ERROR)
                RNS.log("The contained exception was: %s" % (str(e)), RNS.LOG_ERROR)
                nomadnet.panic()

        self.ignored_list = []
        if os.path.isfile(self.ignoredpath):
            try:
                fh = open(self.ignoredpath, "rb")
                ignored_input = fh.read()
                fh.close()

                ignored_hash_strs = ignored_input.splitlines()

                for hash_str in ignored_hash_strs:
                    if len(hash_str) == RNS.Identity.TRUNCATED_HASHLENGTH//8*2:
                        try:
                            ignored_hash = bytes.fromhex(hash_str.decode("utf-8"))
                            self.ignored_list.append(ignored_hash)

                        except Exception as e:
                            RNS.log("Could not decode RNS Identity hash from: "+str(hash_str), RNS.LOG_DEBUG)
                            RNS.log("The contained exception was: "+str(e), RNS.LOG_DEBUG)

            except Exception as e:
                RNS.log("Error while loading list of ignored destinations: "+str(e), RNS.LOG_ERROR)

        self.directory = nomadnet.Directory(self)

        from nomadnet.RRC import RRCManager
        self.rrc = RRCManager(self)
        self.rrc.load()

        static_peers = []
        for static_peer in self.static_peers:
            try:
                dh = bytes.fromhex(static_peer)
                if len(dh) != RNS.Reticulum.TRUNCATED_HASHLENGTH//8:
                    raise ValueError("Invalid destination length")
                static_peers.append(dh)
            except Exception as e:
                RNS.log(f"Could not decode static peer destination hash {static_peer}: {e}", RNS.LOG_ERROR)

        self.message_router = LXMF.LXMRouter(
            identity = self.identity, storagepath = self.storagepath, autopeer = True,
            propagation_limit = self.lxmf_max_propagation_size, sync_limit = self.lxmf_max_sync_size, delivery_limit = self.lxmf_max_incoming_size,
            max_peers = self.max_peers, static_peers = static_peers, propagation_cost=self.node_propagation_cost
        )

        self.message_router.register_delivery_callback(self.lxmf_delivery)

        for destination_hash in self.ignored_list:
            self.message_router.ignore_destination(destination_hash)

        self.lxmf_destination = self.message_router.register_delivery_identity(self.identity, display_name=self.peer_settings["display_name"], stamp_cost=self.required_stamp_cost)
        if not self.accept_invalid_stamps:
            self.message_router.enforce_stamps()

        RNS.Identity.remember(
            packet_hash=None,
            destination_hash=self.lxmf_destination.hash,
            public_key=self.identity.get_public_key(),
            app_data=None
        )

        RNS.log("LXMF Router ready to receive on: "+RNS.prettyhexrep(self.lxmf_destination.hash))

        if self.enable_node:
            self.message_router.set_message_storage_limit(megabytes=self.message_storage_limit)
            for dest_str in self.prioritised_lxmf_destinations:
                try:
                    dest_hash = bytes.fromhex(dest_str)
                    if len(dest_hash) == RNS.Reticulum.TRUNCATED_HASHLENGTH//8:
                        self.message_router.prioritise(dest_hash)

                except Exception as e:
                    RNS.log("Cannot prioritise "+str(dest_str)+", it is not a valid destination hash", RNS.LOG_ERROR)

            if self.disable_propagation:
                if os.path.isfile(self.pnannouncedpath):
                    try:
                        RNS.log("Sending indication to peered LXMF Propagation Node that this node is no longer participating", RNS.LOG_DEBUG)
                        self.message_router.disable_propagation()
                        os.unlink(self.pnannouncedpath)
                    except Exception as e:
                        RNS.log("An error ocurred while indicating that this LXMF Propagation Node is no longer participating. The contained exception was: "+str(e), RNS.LOG_ERROR)
            else:
                self.message_router.enable_propagation()
                try:
                    with open(self.pnannouncedpath, "wb") as pnf:
                        pnf.write(msgpack.packb(time.time()))
                        pnf.close()

                except Exception as e:
                    RNS.log("An error ocurred while writing Propagation Node announce timestamp. The contained exception was: "+str(e), RNS.LOG_ERROR)

            if not self.disable_propagation:
                RNS.log("LXMF Propagation Node started on: "+RNS.prettyhexrep(self.message_router.propagation_destination.hash))

            self.node = nomadnet.Node(self)
        else:
            self.node = None
            if os.path.isfile(self.pnannouncedpath):
                try:
                    RNS.log("Sending indication to peered LXMF Propagation Node that this node is no longer participating", RNS.LOG_DEBUG)
                    self.message_router.disable_propagation()
                    os.unlink(self.pnannouncedpath)
                except Exception as e:
                    RNS.log("An error ocurred while indicating that this LXMF Propagation Node is no longer participating. The contained exception was: "+str(e), RNS.LOG_ERROR)

        RNS.Transport.register_announce_handler(nomadnet.Conversation)
        RNS.Transport.register_announce_handler(nomadnet.Directory)

        self.autoselect_propagation_node()

        if self.peer_announce_at_start:
            def delayed_announce():
                time.sleep(NomadNetworkApp.START_ANNOUNCE_DELAY)
                self.announce_now()

            da_thread = threading.Thread(target=delayed_announce)
            da_thread.setDaemon(True)
            da_thread.start()

        atexit.register(self.exit_handler)
        sys.excepthook = self.exception_handler

        job_thread = threading.Thread(target=self.__jobs)
        job_thread.setDaemon(True)
        job_thread.start()

        # Override UI choice from config on --daemon switch
        if daemon:
            self.uimode = nomadnet.ui.UI_NONE

        # This stderr redirect is needed to stop urwid
        # from spewing KeyErrors to the console and thus,
        # messing up the UI. A pull request to fix the
        # bug in urwid was submitted, but until it is
        # merged, this hack will mitigate it.
        strio = io.StringIO()
        with contextlib.redirect_stderr(strio):
            nomadnet.ui.spawn(self.uimode)

        if strio.tell() > 0:
            try:
                strio.seek(0)
                err_file = open(self.errorfilepath, "w")
                err_file.write(strio.read())
                err_file.close()

            except Exception as e:
                RNS.log("Could not write stderr output to error log file at "+str(self.errorfilepath)+".", RNS.LOG_ERROR)
                RNS.log("The contained exception was: "+str(e), RNS.LOG_ERROR)


    def __jobs(self):
        RNS.log("Deferring scheduled jobs for "+str(self.defer_jobs)+" seconds...", RNS.LOG_DEBUG)
        time.sleep(self.defer_jobs)

        RNS.log("Starting job scheduler now", RNS.LOG_DEBUG)
        while self.should_run_jobs:
            now = time.time()

            if now > self.peer_settings["last_lxmf_sync"] + self.lxmf_sync_interval:
                RNS.log("Initiating automatic LXMF sync", RNS.LOG_VERBOSE)
                self.request_lxmf_sync(limit=self.lxmf_sync_limit)

            if now > self.peer_settings["last_announce"] + self.peer_settings["announce_interval"]:
                self.announce_now()

            time.sleep(self.job_interval)

    def set_display_name(self, display_name):
        self.peer_settings["display_name"] = display_name
        self.lxmf_destination.display_name = display_name
        self.save_peer_settings()

    def get_display_name(self):
        return self.peer_settings["display_name"]

    def get_display_name_bytes(self):
        return self.peer_settings["display_name"].encode("utf-8")

    def get_sync_status(self):
        if self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_IDLE:
            return "Idle"
        elif self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_PATH_REQUESTED:
            return "Path requested"
        elif self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_LINK_ESTABLISHING:
            return "Establishing link"
        elif self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_LINK_ESTABLISHED:
            return "Link established"
        elif self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_REQUEST_SENT:
            return "Sync request sent"
        elif self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_RECEIVING:
            return "Receiving messages"
        elif self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_RESPONSE_RECEIVED:
            return "Messages received"
        elif self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_NO_PATH:
            return "No path to node"
        elif self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_LINK_FAILED:
            return "Link establisment failed"
        elif self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_TRANSFER_FAILED:
            return "Sync request failed"
        elif self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_NO_IDENTITY_RCVD:
            return "Remote got no identity"
        elif self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_NO_ACCESS:
            return "Node rejected request"
        elif self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_FAILED:
            return "Sync failed"
        elif self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_COMPLETE:
            new_msgs = self.message_router.propagation_transfer_last_result
            if new_msgs == 0:
                return "Done, no new messages"
            else:
                return "Downloaded "+str(new_msgs)+" new messages"
        else:
            return "Unknown"

    def sync_status_show_percent(self):
        if self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_IDLE:
            return False
        elif self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_PATH_REQUESTED:
            return False
        elif self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_LINK_ESTABLISHING:
            return False
        elif self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_LINK_ESTABLISHED:
            return False
        elif self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_REQUEST_SENT:
            return False
        elif self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_RECEIVING:
            return True
        elif self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_RESPONSE_RECEIVED:
            return True
        elif self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_COMPLETE:
            return False
        else:
            return False

    def get_sync_progress(self):
        return self.message_router.propagation_transfer_progress

    def request_lxmf_sync(self, limit = None):
        if self.message_router.propagation_transfer_state == LXMF.LXMRouter.PR_IDLE or self.message_router.propagation_transfer_state >= LXMF.LXMRouter.PR_COMPLETE:
            self.peer_settings["last_lxmf_sync"] = time.time()
            self.save_peer_settings()
            self.message_router.request_messages_from_propagation_node(self.identity, max_messages = limit)

    def cancel_lxmf_sync(self):
        if self.message_router.propagation_transfer_state != LXMF.LXMRouter.PR_IDLE:
            self.message_router.cancel_propagation_node_requests()

    def _persist_ignored_list(self):
        try:
            with open(self.ignoredpath, "wb") as fh:
                for h in self.ignored_list:
                    fh.write((RNS.hexrep(h, delimit=False)+"\n").encode("utf-8"))
        except Exception as e:
            RNS.log("Could not persist ignored list: "+str(e), RNS.LOG_ERROR)

    def block_destination(self, dest_hash, reason=None):
        if not isinstance(dest_hash, (bytes, bytearray)):
            return False
        dest_hash = bytes(dest_hash)
        try:
            identity = RNS.Identity.recall(dest_hash)
            if identity is not None:
                RNS.Transport.blackhole_identity(identity.hash, reason=reason)
        except Exception as e:
            RNS.log("Could not blackhole identity: "+str(e), RNS.LOG_ERROR)
        if dest_hash not in self.ignored_list:
            self.ignored_list.append(dest_hash)
            self._persist_ignored_list()
        try:
            self.message_router.ignore_destination(dest_hash)
        except Exception:
            pass
        return True

    def unblock_destination(self, dest_hash):
        if not isinstance(dest_hash, (bytes, bytearray)):
            return False
        dest_hash = bytes(dest_hash)
        try:
            identity = RNS.Identity.recall(dest_hash)
            if identity is not None:
                RNS.Transport.unblackhole_identity(identity.hash)
        except Exception as e:
            RNS.log("Could not lift blackhole on identity: "+str(e), RNS.LOG_ERROR)
        if dest_hash in self.ignored_list:
            self.ignored_list.remove(dest_hash)
            self._persist_ignored_list()
        try:
            self.message_router.unignore_destination(dest_hash)
        except Exception:
            pass
        return True

    def announce_now(self):
        RNS.log("Sending lxmf.delivery announce", RNS.LOG_VERBOSE)
        self.message_router.set_inbound_stamp_cost(self.lxmf_destination.hash, self.required_stamp_cost)
        self.lxmf_destination.display_name = self.peer_settings["display_name"]
        self.message_router.announce(self.lxmf_destination.hash)
        self.peer_settings["last_announce"] = time.time()
        self.save_peer_settings()

    def autoselect_propagation_node(self):
        selected_node = None

        if "propagation_node" in self.peer_settings and self.peer_settings["propagation_node"] != None:
            selected_node = self.peer_settings["propagation_node"]
        else:
            nodes = self.directory.known_nodes()
            trusted_nodes = []

            best_hops = RNS.Transport.PATHFINDER_M+1

            for node in nodes:
                if node.trust_level == DirectoryEntry.TRUSTED:
                    hops = RNS.Transport.hops_to(node.source_hash)

                    if hops < best_hops:
                        best_hops = hops
                        selected_node = node.source_hash

        if selected_node == None:
            RNS.log("Could not autoselect a propagation node! LXMF propagation will not be available until a trusted node announces on the network, or a propagation node is manually selected.", RNS.LOG_WARNING)
        else:
            pn_name_str = ""
            RNS.log("Selecting "+RNS.prettyhexrep(selected_node)+pn_name_str+" as default LXMF propagation node", RNS.LOG_DEBUG)
            self.message_router.set_outbound_propagation_node(selected_node)

    def get_user_selected_propagation_node(self):
        if "propagation_node" in self.peer_settings:
            return self.peer_settings["propagation_node"]
        else:
            return None

    def set_user_selected_propagation_node(self, node_hash):
        self.peer_settings["propagation_node"] = node_hash
        self.save_peer_settings()
        self.autoselect_propagation_node()

    def get_default_propagation_node(self):
        return self.message_router.get_outbound_propagation_node()

    def save_peer_settings(self):
        tmp_path = f"{self.peersettingspath}.tmp"
        with open(tmp_path, "wb") as file: file.write(msgpack.packb(self.peer_settings))
        os.replace(tmp_path, self.peersettingspath)

    def lxmf_delivery(self, message):
        time_string = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(message.timestamp))
        signature_string = "Signature is invalid, reason undetermined"
        if message.signature_validated:
            signature_string = "Validated"
        else:
            if message.unverified_reason == LXMF.LXMessage.SIGNATURE_INVALID:
                signature_string = "Invalid signature"
            if message.unverified_reason == LXMF.LXMessage.SOURCE_UNKNOWN:
                signature_string = "Cannot verify, source is unknown"

        nomadnet.Conversation.ingest(message, self)

        if self.notify_on_new_message:
            self.notify_message_recieved()

        if self.should_print(message):
            self.print_message(message)

    def should_print(self, message):
        if self.print_messages:
            if self.print_all_messages:
                return True

            else:
                source_hash_text = RNS.hexrep(message.source_hash, delimit=False)

                if self.print_trusted_messages:
                    trust_level = self.directory.trust_level(message.source_hash)
                    if trust_level == DirectoryEntry.TRUSTED:
                        return True

                if type(self.allowed_message_print_destinations) is list:
                    if source_hash_text in self.allowed_message_print_destinations:
                        return True

        return False

    def print_file(self, filename):
        print_command = self.print_command+" "+filename

        try:
            return_code = subprocess.call(shlex.split(print_command), stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

        except Exception as e:
            RNS.log("An error occurred while executing print command: "+str(print_command), RNS.LOG_ERROR)
            RNS.log("The contained exception was: "+str(e), RNS.LOG_ERROR)
            return False

        if return_code == 0:
            RNS.log("Successfully printed "+str(filename)+" using print command: "+print_command, RNS.LOG_DEBUG)
            return True

        else:
            RNS.log("Printing "+str(filename)+" failed using print command: "+print_command, RNS.LOG_DEBUG)
            return False


    def print_message(self, message, received = None):
        try:
            template = self.printing_template_msg

            if received == None:
                received = time.time()

            g = self.ui.glyphs

            m_rtime = datetime.fromtimestamp(message.timestamp)
            stime = m_rtime.strftime(self.time_format)

            message_time = datetime.fromtimestamp(received)
            rtime = message_time.strftime(self.time_format)

            display_name = self.directory.simplest_display_str(message.source_hash)
            title = message.title_as_string()
            if title == "":
                title = "None"

            output = template.format(
                origin=display_name,
                stime=stime,
                rtime=rtime,
                mtitle=title,
                mbody=message.content_as_string(),
            )

            filename = "/tmp/"+RNS.hexrep(RNS.Identity.full_hash(output.encode("utf-8")), delimit=False)
            with open(filename, "wb") as f:
                f.write(output.encode("utf-8"))
                f.close()

            self.print_file(filename)

            os.unlink(filename)

        except Exception as e:
            RNS.log("Error while printing incoming LXMF message. The contained exception was: "+str(e))

    def conversations(self):
        return nomadnet.Conversation.conversation_list(self)

    def has_unread_conversations(self):
        if len(nomadnet.Conversation.unread_conversations) > 0:
            return True
        else:
            return False

    def conversation_is_unread(self, source_hash):
        src_bytes = bytes.fromhex(source_hash)
        if src_bytes in nomadnet.Conversation.unread_conversations:
            return True
        if src_bytes in nomadnet.Conversation.failed_conversations:
            return True
        return False

    def mark_conversation_read(self, source_hash):
        src_bytes = bytes.fromhex(source_hash)
        if src_bytes in nomadnet.Conversation.unread_conversations:
            nomadnet.Conversation.unread_conversations.pop(src_bytes)
            if os.path.isfile(self.conversationpath + "/" + source_hash + "/unread"):
                os.unlink(self.conversationpath + "/" + source_hash + "/unread")
        if src_bytes in nomadnet.Conversation.failed_conversations:
            nomadnet.Conversation.failed_conversations.pop(src_bytes)
            if os.path.isfile(self.conversationpath + "/" + source_hash + "/failed"):
                try: os.unlink(self.conversationpath + "/" + source_hash + "/failed")
                except Exception: pass

    def notify_message_recieved(self):
        if self.uimode == nomadnet.ui.UI_TEXT:
            sys.stdout.write("\a")
            sys.stdout.flush()

    def clear_tmp_dir(self):
        if os.path.isdir(self.tmpfilespath):
            for file in os.listdir(self.tmpfilespath):
                fpath = self.tmpfilespath+"/"+file
                os.unlink(fpath)

    def createDefaultConfig(self):
        self.config = ConfigObj(__default_nomadnet_config__)
        self.config.filename = self.configpath

        if not os.path.isdir(self.configdir):
            os.makedirs(self.configdir)
        self.config.write()
        self.applyConfig()


    def applyConfig(self):
        if "logging" in self.config:
            for option in self.config["logging"]:
                value = self.config["logging"][option]
                if option == "loglevel":
                    RNS.loglevel = int(value)
                    if RNS.loglevel < 0:
                        RNS.loglevel = 0
                    if RNS.loglevel > 7:
                        RNS.loglevel = 7
                if option == "destination":
                    if value.lower() == "file" and not self.force_console_log:
                        RNS.logdest = RNS.LOG_FILE
                        if "logfile" in self.config["logging"]:
                            self.logfilepath = self.config["logging"]["logfile"]
                        RNS.logfile = self.logfilepath
                    else:
                        RNS.logdest = RNS.LOG_STDOUT

        if "client" in self.config:
            for option in self.config["client"]:
                value = self.config["client"][option]

                if option == "enable_client":
                    value = self.config["client"].as_bool(option)
                    self.enable_client = value

                if option == "downloads_path":
                    value = self.config["client"]["downloads_path"]
                    self.downloads_path = os.path.expanduser(value)

                if option == "attachment_save_path":
                    value = self.config["client"]["attachment_save_path"]
                    self.attachment_save_path = os.path.expanduser(value)

                if option == "announce_at_start":
                    value = self.config["client"].as_bool(option)
                    self.peer_announce_at_start = value

                if option == "announce_interval":
                    value = self.config["client"].as_int(option)
                    if value < 30: value = 30
                    self.announce_interval = value*60

                if option == "try_propagation_on_send_fail":
                    value = self.config["client"].as_bool(option)
                    self.try_propagation_on_fail = value

                if option == "periodic_lxmf_sync":
                    value = self.config["client"].as_bool(option)
                    self.periodic_lxmf_sync = value

                if option == "lxmf_sync_interval":
                    value = self.config["client"].as_int(option)*60

                    if value >= 60:
                        self.lxmf_sync_interval = value

                if option == "lxmf_sync_limit":
                    value = self.config["client"].as_int(option)

                    if value > 0:
                        self.lxmf_sync_limit = value
                    else:
                        self.lxmf_sync_limit = None

                if option == "required_stamp_cost":
                    value = self.config["client"][option]
                    if value.lower() == "none":
                        self.required_stamp_cost = None
                    else:
                        value = self.config["client"].as_int(option)

                        if value > 0:
                            if value > 255:
                                value = 255
                            self.required_stamp_cost = value
                        else:
                            self.required_stamp_cost = None

                if option == "accept_invalid_stamps":
                    value = self.config["client"].as_bool(option)
                    self.accept_invalid_stamps = value

                if option == "max_accepted_size":
                    value = self.config["client"].as_float(option)

                    if value > 0:
                        self.lxmf_max_incoming_size = value
                    else:
                        self.lxmf_max_incoming_size = 500

                if option == "compact_announce_stream":
                    value = self.config["client"].as_bool(option)
                    self.compact_stream = value

                if option == "notify_on_new_message":
                    value = self.config["client"].as_bool(option)
                    self.notify_on_new_message = value

                if option == "compose_in_markdown":
                    value = self.config["client"].as_bool(option)
                    self.compose_markdown = value

                if option == "user_interface":
                    value = value.lower()
                    if value == "none":
                        self.uimode = nomadnet.ui.UI_NONE
                    if value == "menu":
                        self.uimode = nomadnet.ui.UI_MENU
                    if value == "text":
                        self.uimode = nomadnet.ui.UI_TEXT
                        if "textui" in self.config:
                            if not "intro_time" in self.config["textui"]:
                                self.config["textui"]["intro_time"] = 1
                            else:
                                self.config["textui"]["intro_time"] = self.config["textui"].as_float("intro_time")

                            if not "intro_text" in self.config["textui"]:
                                self.config["textui"]["intro_text"] = "Nomad Network"

                            if not "editor" in self.config["textui"]:
                                self.config["textui"]["editor"] = "nano"

                            if not "glyphs" in self.config["textui"]:
                                self.config["textui"]["glyphs"] = "unicode"

                            if not "mouse_enabled" in self.config["textui"]:
                                self.config["textui"]["mouse_enabled"] = True
                            else:
                                self.config["textui"]["mouse_enabled"] = self.config["textui"].as_bool("mouse_enabled")

                            if not "hide_guide" in self.config["textui"]:
                                self.config["textui"]["hide_guide"] = False
                            else:
                                self.config["textui"]["hide_guide"] = self.config["textui"].as_bool("hide_guide")

                            if not "sanitize_names" in self.config["textui"]:
                                self.config["textui"]["sanitize_names"] = True
                            else:
                                self.config["textui"]["sanitize_names"] = self.config["textui"].as_bool("sanitize_names")

                            if not "clipboard_copy" in self.config["textui"]:
                                self.config["textui"]["clipboard_copy"] = False
                            else:
                                self.config["textui"]["clipboard_copy"] = self.config["textui"].as_bool("clipboard_copy")

                            if not "animation_interval" in self.config["textui"]:
                                self.config["textui"]["animation_interval"] = 1
                            else:
                                self.config["textui"]["animation_interval"] = self.config["textui"].as_int("animation_interval")

                            if not "colormode" in self.config["textui"]:
                                self.config["textui"]["colormode"] = nomadnet.ui.COLORMODE_16
                            else:
                                if self.config["textui"]["colormode"].lower() == "monochrome":
                                    self.config["textui"]["colormode"] = nomadnet.ui.TextUI.COLORMODE_MONO
                                elif self.config["textui"]["colormode"].lower() == "16":
                                    self.config["textui"]["colormode"] = nomadnet.ui.TextUI.COLORMODE_16
                                elif self.config["textui"]["colormode"].lower() == "88":
                                    self.config["textui"]["colormode"] = nomadnet.ui.TextUI.COLORMODE_88
                                elif self.config["textui"]["colormode"].lower() == "256":
                                    self.config["textui"]["colormode"] = nomadnet.ui.TextUI.COLORMODE_256
                                elif self.config["textui"]["colormode"].lower() == "24bit":
                                    self.config["textui"]["colormode"] = nomadnet.ui.TextUI.COLORMODE_TRUE
                                else:
                                    raise ValueError("The selected Text UI color mode is invalid")

                            if not "theme" in self.config["textui"]:
                                self.config["textui"]["theme"] = nomadnet.ui.TextUI.THEME_DARK
                            else:
                                if self.config["textui"]["theme"].lower() == "dark":
                                    self.config["textui"]["theme"] = nomadnet.ui.TextUI.THEME_DARK
                                elif self.config["textui"]["theme"].lower() == "light":
                                    self.config["textui"]["theme"] = nomadnet.ui.TextUI.THEME_LIGHT
                                else:
                                    raise ValueError("The selected Text UI theme is invalid")
                        else:
                            raise KeyError("Text UI selected in configuration file, but no [textui] section found")
                    if value == "graphical":
                        self.uimode = nomadnet.ui.UI_GRAPHICAL
                    if value == "web":
                        self.uimode = nomadnet.ui.UI_WEB

        if "rrc" in self.config:
            for option in self.config["rrc"]:
                if option == "history_per_room_cap":
                    try: value = self.config["rrc"].as_int(option)
                    except Exception: value = None
                    if value is not None and value >= 0:
                        self.rrc_history_per_room_cap = value

                if option == "filter_loaded_history":
                    try: value = self.config["rrc"].as_bool(option)
                    except Exception: value = True
                    self.rrc_filter_loaded_history = value

                if option == "ephemeral_notices":
                    try: value = self.config["rrc"].as_float(option)
                    except Exception: value = 0
                    self.rrc_ephemeral_notices = value*60

                if option == "justify_msgs":
                    try: value = self.config["rrc"].as_bool(option)
                    except Exception: value = True
                    self.rrc_ui_justify_msgs = value

                if option == "space_msgs":
                    try: value = self.config["rrc"].as_bool(option)
                    except Exception: value = False
                    self.rrc_ui_space_msgs = value

                if option == "nick_colors":
                    try: value = self.config["rrc"].as_bool(option)
                    except Exception: value = True
                    self.rrc_nick_colors = value

                if option == "mention_color":
                    value = self.config["rrc"][option]
                    if len(value) == 6:
                        try: bytes.fromhex(value)
                        except Exception: value = None
                        self.rrc_mention_color = value

                if option == "color_mention_timestamps":
                    try: value = self.config["rrc"].as_bool(option)
                    except Exception: value = True
                    self.rrc_color_mention_timestamps = value

                if option == "nick_colors_theme":
                    try:
                        colors = self.config["rrc"].as_list(option)
                        nick_colors_theme = []
                        for c in colors:
                            if len(c) == 6:
                                try: bytes.fromhex(c)
                                except: continue
                                nick_colors_theme.append(c)
                        self.rrc_nick_colors_theme = nick_colors_theme
                    except Exception as e:
                        RNS.log(f"Could not load custom nick colors theme: {e}", RNS.LOG_WARNING)
                        self.rrc_nick_colors_theme = None

                if option == "render_markdown":
                    try: value = self.config["rrc"].as_bool(option)
                    except Exception: value = True
                    self.rrc_ui_render_markdown = value

                if option == "render_micron":
                    try: value = self.config["rrc"].as_bool(option)
                    except Exception: value = True
                    self.rrc_ui_render_micron = value

                if option == "show_gutters":
                    try: value = self.config["rrc"].as_bool(option)
                    except Exception: value = False
                    self.rrc_show_gutters = value

                if option == "enable_esoterics":
                    try: value = self.config["rrc"].as_bool(option)
                    except Exception: value = False
                    self.rrc_enable_esoterics = value

        if "node" in self.config:
            if not "enable_node" in self.config["node"]:
                self.enable_node = False
            else:
                self.enable_node = self.config["node"].as_bool("enable_node")

            if not "node_name" in self.config["node"]:
                self.node_name = None
            else:
                value = self.config["node"]["node_name"]
                if value.lower() == "none":
                    self.node_name = None
                else:
                    self.node_name = self.config["node"]["node_name"]

            if not "disable_propagation" in self.config["node"]:
                self.disable_propagation = True
            else:
                self.disable_propagation = self.config["node"].as_bool("disable_propagation")

            if not "max_transfer_size" in self.config["node"]:
                self.lxmf_max_propagation_size = 256
            else:
                value = self.config["node"].as_float("max_transfer_size")
                if value < 1:
                    value = 1
                self.lxmf_max_propagation_size = value

            if not "max_sync_size" in self.config["node"]:
                self.lxmf_max_sync_size = 256*40
            else:
                value = self.config["node"].as_float("max_sync_size")
                if value < self.lxmf_max_propagation_size:
                    value = self.lxmf_max_propagation_size
                self.lxmf_max_sync_size = value

            if not "announce_at_start" in self.config["node"]:
                self.node_announce_at_start = False
            else:
                value = self.config["node"].as_bool("announce_at_start")
                self.node_announce_at_start = value

            if not "announce_interval" in self.config["node"]:
                self.node_announce_interval = 720
            else:
                value = self.config["node"].as_int("announce_interval")
                if value < 1:
                    value = 1
                self.node_announce_interval = value

            if not "propagation_cost" in self.config["node"]:
                self.node_propagation_cost = 16
            else:
                value = self.config["node"].as_int("propagation_cost")
                if value < 13: value = 13
                self.node_propagation_cost = value

            if "pages_path" in self.config["node"]:
                self.pagespath = self.config["node"]["pages_path"]

            if not "page_refresh_interval" in self.config["node"]:
                self.page_refresh_interval = 0
            else:
                value = self.config["node"].as_int("page_refresh_interval")
                if value < 0:
                    value = 0
                self.page_refresh_interval = value


            if "files_path" in self.config["node"]:
                self.filespath = self.config["node"]["files_path"]

            if not "file_refresh_interval" in self.config["node"]:
                self.file_refresh_interval = 0
            else:
                value = self.config["node"].as_int("file_refresh_interval")
                if value < 0:
                    value = 0
                self.file_refresh_interval = value


            if "prioritise_destinations" in self.config["node"]:
                self.prioritised_lxmf_destinations = self.config["node"].as_list("prioritise_destinations")
            else:
                self.prioritised_lxmf_destinations = []

            if "static_peers" in self.config["node"]:
                self.static_peers = self.config["node"].as_list("static_peers")
            else:
                self.static_peers = []

            if not "max_peers" in self.config["node"]:
                self.max_peers = None
            else:
                value = self.config["node"].as_int("max_peers")
                if value < 0:
                    value = 0
                self.max_peers = value

            if not "message_storage_limit" in self.config["node"]:
                self.message_storage_limit = 2000
            else:
                value = self.config["node"].as_float("message_storage_limit")
                if value < 0.005:
                    value = 0.005
                self.message_storage_limit = value

        self.print_command = "lp"
        self.print_messages = False
        self.print_all_messages = False
        self.print_trusted_messages = False
        if "printing" in self.config:
            if not "print_messages" in self.config["printing"]:
                self.print_messages = False
            else:
                self.print_messages = self.config["printing"].as_bool("print_messages")

                if "print_command" in self.config["printing"]:
                    self.print_command = self.config["printing"]["print_command"]

                if self.print_messages:
                    if not "print_from" in self.config["printing"]:
                        self.allowed_message_print_destinations = None
                    else:
                        if type(self.config["printing"]["print_from"]) == str:
                            self.allowed_message_print_destinations = []
                            if self.config["printing"]["print_from"].lower() == "everywhere":
                                self.print_all_messages = True

                            if self.config["printing"]["print_from"].lower() == "trusted":

                                self.print_all_messages = False
                                self.print_trusted_messages = True

                            if len(self.config["printing"]["print_from"]) == (RNS.Identity.TRUNCATED_HASHLENGTH//8)*2:
                                self.allowed_message_print_destinations.append(self.config["printing"]["print_from"])

                        if type(self.config["printing"]["print_from"]) == list:
                                self.allowed_message_print_destinations =  self.config["printing"].as_list("print_from")
                                for allowed_entry in self.allowed_message_print_destinations:
                                    if allowed_entry.lower() == "trusted":
                                        self.print_trusted_messages = True


                    if not "message_template" in self.config["printing"]:
                        self.printing_template_msg = __printing_template_msg__
                    else:
                        mt_path = os.path.expanduser(self.config["printing"]["message_template"])
                        if os.path.isfile(mt_path):
                            template_file = open(mt_path, "rb")
                            self.printing_template_msg = template_file.read().decode("utf-8")
                        else:
                            template_file = open(mt_path, "wb")
                            template_file.write(__printing_template_msg__.encode("utf-8"))
                            self.printing_template_msg = __printing_template_msg__


    @staticmethod
    def get_shared_instance():
        if NomadNetworkApp._shared_instance != None:
            return NomadNetworkApp._shared_instance
        else:
            raise UnboundLocalError("No Nomad Network applications have been instantiated yet")


    def quit(self):
        RNS.log("Nomad Network Client shutting down...")
        os._exit(0)


# Default configuration file:
__default_nomadnet_config__ = '''# This is the default Nomad Network config file.
# You should probably edit it to suit your needs and use-case,

[logging]
# Valid log levels are 0 through 7:
#   0: Log only critical information
#   1: Log errors and lower log levels
#   2: Log warnings and lower log levels
#   3: Log notices and lower log levels
#   4: Log info and lower (this is the default)
#   5: Verbose logging
#   6: Debug logging
#   7: Extreme logging

loglevel = 4
destination = file

[client]

enable_client = yes
user_interface = text
downloads_path = ~/Downloads

# Where to save received attachments. If not set,
# attachments will be saved to the downloads path.
# attachment_save_path = ~/Downloads
notify_on_new_message = yes

# By default, the peer is announced at startup
# to let other peers reach it immediately.
announce_at_start = yes

# Automatic announce interval in minutes for
# your LXMF address. 6 hours by default.
announce_interval = 360

# By default, the client will try to deliver a
# message via the LXMF propagation network, if
# a direct delivery to the recipient is not
# possible.
try_propagation_on_send_fail = yes

# Nomadnet will periodically sync messages from
# LXMF propagation nodes by default, if any are
# present. You can disable this if you want to
# only sync when manually initiated.
periodic_lxmf_sync = yes

# The sync interval in minutes. This value is
# equal to 6 hours (360 minutes) by default.
lxmf_sync_interval = 360

# By default, automatic LXMF syncs will only
# download 8 messages at a time. You can change
# this number, or set the option to 0 to disable
# the limit, and download everything every time.
lxmf_sync_limit = 8

# You can specify a required stamp cost for
# inbound messages to be accepted. Specifying
# a stamp cost will require untrusted senders
# that message you to include a cryptographic
# stamp in their messages. Performing this
# operation takes the sender an amount of time
# proportional to the stamp cost. As a rough
# estimate, a stamp cost of 8 will take less
# than a second to compute, and a stamp cost
# of 20 could take several minutes, even on
# a fast computer.
required_stamp_cost = None

# You can signal stamp requirements to senders,
# but still accept messages with invalid stamps
# by setting this option to True.
accept_invalid_stamps = False

# The maximum accepted unpacked size for mes-
# sages received directly from other peers,
# specified in kilobytes. Messages larger than
# this will be rejected before the transfer
# begins.
max_accepted_size = 500

# The announce stream will only show one entry
# per destination or node by default. You can
# change this to show as many announces as have
# been received, for every destination.
compact_announce_stream = yes

# You can choose whether or not to compose
# messages in markdown format.
compose_in_markdown = yes

[textui]

# Amount of time to show intro screen
intro_time = 1

# You can specify the display theme.
# theme = light
theme = dark

# Specify the number of colors to use
# valid colormodes are:
# monochrome, 16, 88, 256 and 24bit
#
# The default is a conservative 256 colors.
# If your terminal does not support this,
# you can lower it. Some terminals support
# 24 bit color.

# colormode = monochrome
# colormode = 16
# colormode = 88
# colormode = 256
colormode = 24bit

# By default, unicode glyphs are used. If
# you have a Nerd Font installed, you can
# enable this for a better user interface.
# You can also enable plain text glyphs if
# your terminal doesn't support unicode.

# glyphs = plain
# glyphs = unicode
glyphs = nerdfont

# You can specify whether mouse events
# should be considered as input to the
# application. On by default.
mouse_enabled = True

# What editor to use for editing text.
editor = nano

# If you don't want the Guide section to
# show up in the menu, you can disable it.
hide_guide = no

# You can select whether or not to perform
# text sanitization on names received in
# announces.
sanitize_names = yes

# You can enable clipboard features. This
# relies on your terminal supporting OSC52
# escape sequences.
clipboard_copy = no

[rrc]

# Maximum number of messages retained per
# room in the in-memory scrollback buffer,
# and the number of messages restored from
# on-disk history at startup. The on-disk
# log itself is appended to indefinitely;
# this cap only controls how much backlog
# is visible. Set to 0 to keep every message
# in memory (full history).
history_per_room_cap = 500

# You can choose whether to filter system
# and notice events when room history is
# loaded.
filter_loaded_history = yes

# You can choose whether notices and sys
# messages persist indefinitely, or are
# removed from the message history after
# the specified time in minutes. Set to 0
# to disable and keep forever.
ephemeral_notices = 10

# Other display and formatting options:
color_mention_timestamps = yes
render_markdown = yes
render_micron = yes
nick_colors = yes
justify_msgs = yes
space_msgs = no
show_gutters = yes

# You can configure your own color theme
# for nick color assignment
# nick_colors_theme = f68787, 00c394, d59e00, ...

# You can set a specific color for mentions
# of your nick
# mention_color = FFBB44


[node]

# Whether to enable node hosting
enable_node = no

# The node name will be visible to other
# peers on the network, and included in
# announces.
node_name = None

# Automatic announce interval in minutes.
# 6 hours by default.
announce_interval = 360

# Whether to announce when the node starts.
announce_at_start = Yes

# When Nomad Network is hosting a page-serving
# node, it can also act as an LXMF propagation
# node. This is a convenient feature that lets
# you easily set up and run a propagation node
# on the network, but it is not as fully
# featured as using the lxmd program to host a
# propagation node. For complete control and
# flexibility, use lxmd to run a PN. For a
# small local system or network, the built-in
# PN functionality will suffice for most cases.
#
# If there is already a large amount of
# propagation nodes on the network, or you
# simply want to run a pageserving-only node,
# you should disable running a propagation node.
# Due to lots of propagation nodes being
# available, this is currently the default.

disable_propagation = Yes

# For clients and other propagation nodes
# delivering messages via this node, you can
# configure the minimum required propagation
# stamp costs. All messages delivered to the
# propagation node network must have a valid
# propagation stamp, or they will be rejected.
# Clients automatically detect the stamp cost
# for the node they are delivering to, and
# compute a corresponding stamp before trying
# to deliver the message to the propagation
# node.
#
# Propagation stamps are easier to verify in
# large batches, and therefore also somewhat
# easier to compute for the senders. As such,
# a reasonable propagation stamp cost should
# be a bit higher than the normal peer-to-peer
# stamp costs.
#
# Propagation stamps does not incur any extra
# load for propagation nodes processing them,
# since they are only required to verify that
# they are correct, and only the generation
# is computationally costly. Setting a sensible
# propagation stamp cost (and periodically
# checking the average network consensus) helps
# keep spam and misuse out of the propagation
# node network.

propagation_cost = 16

# The maximum amount of storage to use for
# the LXMF Propagation Node message store,
# specified in megabytes. When this limit
# is reached, LXMF will periodically remove
# messages in its message store. By default,
# LXMF prioritises keeping messages that are
# new and small. Large and old messages will
# be removed first. This setting is optional
# and defaults to 2 gigabytes.

# message_storage_limit = 2000

# The maximum accepted transfer size per in-
# coming propagation message, in kilobytes.
# This sets the upper limit for the size of
# single messages accepted onto this node.

max_transfer_size = 256

# The maximum accepted transfer size per in-
# coming propagation node sync.
#
# If a node wants to propagate a larger number
# of messages to this node, than what can fit
# within this limit, it will prioritise sending
# the smallest messages first, and try again
# with any remaining messages at a later point.

max_sync_size = 10240

# You can tell the LXMF message router to
# prioritise storage for one or more
# destinations. If the message store reaches
# the specified limit, LXMF will prioritise
# keeping messages for destinations specified
# with this option. This setting is optional,
# and generally you do not need to use it.

# prioritise_destinations = 41d20c727598a3fbbdf9106133a3a0ed, d924b81822ca24e68e2effea99bcb8cf

# You can configure the maximum number of other
# propagation nodes that this node will peer
# with automatically. The default is 20.

# max_peers = 20

# You can configure a list of static propagation
# node peers, that this node will always be
# peered with, by specifying a list of
# destination hashes.

# static_peers = e17f833c4ddf8890dd3a79a6fea8161d, 5a2d0029b6e5ec87020abaea0d746da4

# You can specify the interval in minutes for
# rescanning the hosted pages path. By default,
# this option is disabled, and the pages path
# will only be scanned on startup.

# page_refresh_interval = 0

# You can specify the interval in minutes for
# rescanning the hosted files path. By default,
# this option is disabled, and the files path
# will only be scanned on startup.

# file_refresh_interval = 0

[printing]

# You can configure Nomad Network to print
# various kinds of information and messages.

# Printing messages is disabled by default

print_messages = No

# You can configure a custom template for
# message printing. If you uncomment this
# option, set a path to the template and
# restart Nomad Network, a default template
# will be created that you can edit.

# message_template = ~/.nomadnetwork/print_template_msg.txt

# You can configure Nomad Network to only
# print messages from trusted destinations.

# print_from = trusted

# Or specify the source LXMF addresses that
# will automatically have messages printed
# on arrival.

# print_from = 76fe5751a56067d1e84eef3e88eab85b, 0e70b5848eb57c13154154feaeeb89b7

# Or allow printing from anywhere, if you
# are feeling brave and adventurous.

# print_from = everywhere

# You can configure the printing command.
# This will use the default CUPS printer on
# your system.

print_command = lp

# You can specify what printer to use
# print_command = lp -d [PRINTER_NAME]

# Or specify more advanced options. This
# example works well for small thermal-
# roll printers:
# print_command = lp -d [PRINTER_NAME] -o cpi=16 -o lpi=8

# This one is more suitable for full-sheet
# printers. It will print a QR code at the center of any media
# your printer will accept, print in portrait mode, and move the message to
# the top of the print queue:
# print_command = lp -d [PRINTER_NAME] -o job-priority=100 -o media=Custom.75x75mm -o orientation-requested=3

# But you can modify the size to fit your needs.
# The custom media option accepts millimeters, centimeters, and
# inches in a width by length format like so:
# -o media=Custom.[WIDTH]x[LENGTH][mm,cm,in]
#
# The job priority option accepts 1-100, though you can remove it
# entirely if you aren't concerned with a print queue:
# -o job-priority=[1-100]
#
# Finally, the orientation option allows for 90 degree rotations beginning with 3, so:
# -o orientation-requested=4 (landscape, 90 degrees)
# -o orientation-requested=5 (reverse portrait, 180 degrees)
#
# Here is the full command with the recommended customizable variables:
# print_command = lp -d [PRINTER_NAME] -o job-priority=[N] -o media=[MEDIA_SIZE] -o orientation-requested=[N] -o sides=one-sided

# For example, here's a configuration for USB thermal printer that uses the POS-58 PPD driver
# with rolls 47.98x209.9mm in size:
# print_command = lp -d [PRINTER_NAME] -o job-priority=100 -o media=custom_47.98x209.9mm_47.98x209.9mm -o sides=one-sided

'''.splitlines()

__printing_template_msg__ = """
---------------------------
From: {origin}
Sent: {stime}
Rcvd: {rtime}
Title: {mtitle}

{mbody}
---------------------------
"""



import RNS
import urwid
import nomadnet
import time
import threading
from datetime import datetime
from nomadnet.Directory import DirectoryEntry
from nomadnet.vendor.additional_urwid_widgets import IndicativeListBox, MODIFIER_KEY
from nomadnet.util import strip_modifiers
from nomadnet.util import sanitize_name

from .Browser import Browser

class NetworkDisplayShortcuts():
    def __init__(self, app):
        self.app = app
        g = app.ui.glyphs

        shortcut_text = "[C-l] Nodes/Announces  [C-x] Remove  [C-w] Disconnect  [C-d] Back  [C-f] Forward  [C-r] Reload  [C-u] URL  "
        if app.config["textui"]["clipboard_copy"]:
            shortcut_text += "[C-y] Copy  "
        shortcut_text += "[C-g] Fullscreen  [C-s / C-b] Save Node"
        self.widget = urwid.AttrMap(urwid.Text(shortcut_text), "shortcutbar")

class DialogLineBox(urwid.LineBox):
    def keypress(self, size, key):
        if key == "esc":
            self.delegate.update_conversation_list()
        else:
            return super(DialogLineBox, self).keypress(size, key)


class ListEntry(urwid.Text):
    _selectable = True

    signals = ["click"]

    def keypress(self, size, key):
        """
        Send 'click' signal on 'activate' command.
        """
        if self._command_map[key] != urwid.ACTIVATE:
            return key

        self._emit('click')

    def mouse_event(self, size, event, button, x, y, focus):
        """
        Send 'click' signal on button 1 press.
        """
        if button != 1 or not urwid.util.is_mouse_press(event):
            return False

        self._emit('click')
        return True


class AnnounceInfo(urwid.WidgetWrap):
    def keypress(self, size, key):
        if key == "esc":
            options = self.parent.left_pile.options(height_type=urwid.WEIGHT, height_amount=1)
            self.parent.left_pile.contents[0] = (self.parent.announce_stream_display, options)
        else:
            return super(AnnounceInfo, self).keypress(size, key)

    def __init__(self, announce, parent, app):
        self.app = nomadnet.NomadNetworkApp.get_shared_instance()
        self.parent = self.app.ui.main_display.sub_displays.network_display
        g = self.app.ui.glyphs

        source_hash  = announce[1]
        time_format  = app.time_format
        dt           = datetime.fromtimestamp(announce[0])
        ts_string    = dt.strftime(time_format)
        trust_level  = self.app.directory.trust_level(source_hash)
        trust_str    = ""
        display_str  = self.app.directory.simplest_display_str(source_hash)
        addr_str     = "<"+RNS.hexrep(source_hash, delimit=False)+">"
        info_type    = announce[3]

        is_node = False
        is_pn   = False
        if info_type == "node" or info_type == True:
            type_string = "Nomad Network Node " + g["node"]
            is_node = True
        elif info_type == "pn":
            type_string = "LXMF Propagation Node " + g["sent"]
            is_pn = True
        elif info_type == "peer" or info_type == False:
            type_string = "Peer " + g["peer"]

        try:
            data_str = strip_modifiers(announce[2].decode("utf-8"))
            data_style = ""
            if trust_level != DirectoryEntry.TRUSTED and len(data_str) > 32:
                data_str = data_str[:32]+" [...]"
        except Exception as e:
            data_str = "Decode failed"
            data_style = "list_untrusted"


        if trust_level == DirectoryEntry.UNTRUSTED:
            trust_str     = "Untrusted"
            symbol        = g["cross"]
            style         = "list_untrusted"
        elif trust_level == DirectoryEntry.UNKNOWN:
            trust_str     = "Unknown"
            symbol        = g["unknown"]
            style         = "list_unknown"
        elif trust_level == DirectoryEntry.TRUSTED:
            trust_str     = "Trusted"
            symbol        = g["check"]
            style         = "list_trusted"
        elif trust_level == DirectoryEntry.WARNING:
            trust_str     = "Warning"
            symbol        = g["warning"]
            style         = "list_warning"
        else:
            trust_str     = "Warning"
            symbol        = g["warning"]
            style         = "list_untrusted"

        def show_announce_stream(sender):
            options = self.parent.left_pile.options(height_type=urwid.WEIGHT, height_amount=1)
            self.parent.left_pile.contents[0] = (self.parent.announce_stream_display, options)

        def connect(sender):
            self.parent.browser.retrieve_url(RNS.hexrep(source_hash, delimit=False))
            show_announce_stream(None)

        def save_node(sender):
            node_entry = DirectoryEntry(source_hash, display_name=data_str, trust_level=trust_level, hosts_node=True)
            self.app.directory.remember(node_entry)
            self.app.ui.main_display.sub_displays.network_display.directory_change_callback()
            show_announce_stream(None)

        if is_node:
            node_ident = RNS.Identity.recall(source_hash)
            if not node_ident:
                raise KeyError("Could not recall identity for selected node")

            op_hash = RNS.Destination.hash_from_name_and_identity("lxmf.delivery", node_ident)
            op_str = self.app.directory.simplest_display_str(op_hash)

        def msg_op(sender):
            show_announce_stream(None)
            if is_node:
                try:
                    existing_conversations = nomadnet.Conversation.conversation_list(self.app)

                    source_hash_text = RNS.hexrep(op_hash, delimit=False)
                    display_name = op_str

                    if not source_hash_text in [c[0] for c in existing_conversations]:
                        entry = DirectoryEntry(source_hash, display_name, trust_level)
                        self.app.directory.remember(entry)

                        new_conversation = nomadnet.Conversation(source_hash_text, nomadnet.NomadNetworkApp.get_shared_instance(), initiator=True)
                        self.app.ui.main_display.sub_displays.conversations_display.update_conversation_list()

                    self.app.ui.main_display.sub_displays.conversations_display.display_conversation(None, source_hash_text)
                    self.app.ui.main_display.show_conversations(None)

                except Exception as e:
                    RNS.log("Error while starting conversation from announce. The contained exception was: "+str(e), RNS.LOG_ERROR)

        def converse(sender):
            show_announce_stream(None)
            try:
                existing_conversations = nomadnet.Conversation.conversation_list(self.app)

                source_hash_text = RNS.hexrep(source_hash, delimit=False)
                display_name = data_str

                if not source_hash_text in [c[0] for c in existing_conversations]:
                    entry = DirectoryEntry(source_hash, display_name, trust_level)
                    self.app.directory.remember(entry)

                    new_conversation = nomadnet.Conversation(source_hash_text, nomadnet.NomadNetworkApp.get_shared_instance(), initiator=True)
                    self.app.ui.main_display.sub_displays.conversations_display.update_conversation_list()

                self.app.ui.main_display.sub_displays.conversations_display.display_conversation(None, source_hash_text)
                self.app.ui.main_display.show_conversations(None)

            except Exception as e:
                RNS.log("Error while starting conversation from announce. The contained exception was: "+str(e), RNS.LOG_ERROR)

        def use_pn(sender):
            show_announce_stream(None)
            try:
                self.app.set_user_selected_propagation_node(source_hash)
            except Exception as e:
                RNS.log("Error while setting active propagation node from announce. The contained exception was: "+str(e), RNS.LOG_ERROR)

        if is_node:
            type_button = (urwid.WEIGHT, 0.45, urwid.Button("Connect", on_press=connect))
            msg_button =  (urwid.WEIGHT, 0.45, urwid.Button("Msg Op", on_press=msg_op))
            save_button = (urwid.WEIGHT, 0.45, urwid.Button("Save", on_press=save_node))
        elif is_pn:
            type_button = (urwid.WEIGHT, 0.45, urwid.Button("Use as default", on_press=use_pn))
            save_button = None
        else:
            type_button = (urwid.WEIGHT, 0.45, urwid.Button("Converse", on_press=converse))
            save_button = None

        if is_node:
            button_columns = urwid.Columns([
                (urwid.WEIGHT, 0.45, urwid.Button("Back", on_press=show_announce_stream)),
                (urwid.WEIGHT, 0.1, urwid.Text("")),
                type_button,
                (urwid.WEIGHT, 0.1, urwid.Text("")),
                msg_button,
                (urwid.WEIGHT, 0.1, urwid.Text("")),
                save_button,
            ])
        else:
            button_columns = urwid.Columns([
                (urwid.WEIGHT, 0.45, urwid.Button("Back", on_press=show_announce_stream)),
                (urwid.WEIGHT, 0.1, urwid.Text("")),
                type_button,
            ])

        pile_widgets = []

        if is_pn:
            pile_widgets = [
                urwid.Text("Time  : "+ts_string, align=urwid.LEFT),
                urwid.Text("Addr  : "+addr_str, align=urwid.LEFT),
                urwid.Text("Type  : "+type_string, align=urwid.LEFT),
                urwid.Divider(g["divider1"]),
                button_columns
            ]

        else:
            pile_widgets = [
                urwid.Text("Time  : "+ts_string, align=urwid.LEFT),
                urwid.Text("Addr  : "+addr_str, align=urwid.LEFT),
                urwid.Text("Type  : "+type_string, align=urwid.LEFT),
                urwid.Text("Name  : "+display_str, align=urwid.LEFT),
                urwid.Text(["Trust : ", (style, trust_str)], align=urwid.LEFT),
                urwid.Divider(g["divider1"]),
                urwid.Text(["Announce Data: \n", (data_style, data_str)], align=urwid.LEFT),
                urwid.Divider(g["divider1"]),
                button_columns
            ]

            if is_node:
                operator_entry = urwid.Text("Oprtr : "+op_str, align=urwid.LEFT)
                pile_widgets.insert(4, operator_entry)

        pile = urwid.Pile(pile_widgets)

        self.display_widget = urwid.Filler(pile, valign=urwid.TOP, height=urwid.PACK)

        super().__init__(urwid.LineBox(self.display_widget, title="Announce Info"))


class AnnounceStreamEntry(urwid.WidgetWrap):
    def __init__(self, app, announce, delegate, show_destination=False):
        full_time_format = "%Y-%m-%d %H:%M:%S"
        date_time_format = "%Y-%m-%d"
        time_time_format = "%H:%M:%S"
        short_time_format = "%Y-%m-%d %H:%M"
        date_only_format = "%Y-%m-%d"

        timestamp = announce[0]
        source_hash = announce[1]
        announce_type = announce[3]
        self.app = app
        self.delegate = delegate
        self.timestamp = timestamp
        time_format = app.time_format
        dt = datetime.fromtimestamp(self.timestamp)
        dtn = datetime.fromtimestamp(time.time())
        g = self.app.ui.glyphs

        if dt.strftime(date_time_format) == dtn.strftime(date_time_format):
            ts_string = dt.strftime(time_time_format)
        else:
            ts_string = dt.strftime(date_only_format)

        trust_level  = self.app.directory.trust_level(source_hash)

        def san(name):
            if self.app.config["textui"]["sanitize_names"]: return sanitize_name(name)
            else:                                           return strip_modifiers(name)
        
        if show_destination:
            display_str = RNS.hexrep(source_hash, delimit=False)
        else:
            try:    display_str = san(announce[2].decode("utf-8"))
            except: display_str = self.app.directory.simplest_display_str(source_hash)

        if not display_str: display_str = RNS.prettyhexrep(source_hash)
        if len(display_str) > 34: display_str = display_str[:34] + "…"

        if trust_level == DirectoryEntry.UNTRUSTED:
            symbol        = g["cross"]
            style         = "list_untrusted"
            focus_style   = "list_focus_untrusted"
        elif trust_level == DirectoryEntry.UNKNOWN:
            symbol        = g["unknown"]
            style         = "list_unknown"
            focus_style   = "list_focus"
        elif trust_level == DirectoryEntry.TRUSTED:
            symbol        = g["check"]
            style         = "list_trusted"
            focus_style   = "list_focus_trusted"
        elif trust_level == DirectoryEntry.WARNING:
            symbol        = g["warning"]
            style         = "list_warning"
            focus_style   = "list_focus"
        else:
            symbol        = g["warning"]
            style         = "list_untrusted"
            focus_style   = "list_focus_untrusted"

        if announce_type == "node" or announce_type == True:
            type_symbol = g["node"]
        elif announce_type == "peer" or announce_type == False:
            type_symbol = g["peer"]
        elif announce_type == "pn":
            type_symbol = g["sent"]

        widget = ListEntry(ts_string+" "+type_symbol+" "+display_str)
        urwid.connect_signal(widget, "click", self.display_announce, announce)

        self.display_widget = urwid.AttrMap(widget, style, focus_style)
        super().__init__(self.display_widget)

    def display_announce(self, event, announce):
      try:
        parent = self.app.ui.main_display.sub_displays.network_display
        info_widget = AnnounceInfo(announce, parent, self.app)
        options = parent.left_pile.options(height_type=urwid.WEIGHT, height_amount=1)
        parent.left_pile.contents[0] = (info_widget, options)

      except KeyError as e:
        def dismiss_dialog(sender):
            self.delegate.parent.close_list_dialogs()

        def confirmed(sender):
            def close_req(sender):
                self.delegate.parent.close_list_dialogs()

            dialog_pile.contents[0] = (urwid.Text("\nKeys requested from network\n", align=urwid.CENTER), options)
            RNS.Transport.request_path(announce[1])

        confirmed_button = urwid.Button("Request keys", on_press=confirmed)

        dialog_pile = urwid.Pile([
            urwid.Text(
                "The keys for the announced destination could not be recalled. "
                "You can wait for an announce to arrive, or request the keys from the network.\n",
                align=urwid.CENTER,
            ),
            urwid.Columns([
                (urwid.WEIGHT, 0.45, confirmed_button),
                (urwid.WEIGHT, 0.1, urwid.Text("")),
                (urwid.WEIGHT, 0.45, urwid.Button("Close", on_press=dismiss_dialog)),
            ])
        ])

        dialog = ListDialogLineBox(
            dialog_pile,
            title="Keys Unknown"
        )
        confirmed_button.dialog_pile = dialog_pile
        dialog.delegate = self.delegate.parent
        bottom = self.delegate

        overlay = urwid.Overlay(
            dialog,
            bottom,
            align=urwid.CENTER,
            width=urwid.RELATIVE_100,
            valign=urwid.MIDDLE,
            height=urwid.PACK,
            left=2,
            right=2,
        )

        options = self.delegate.parent.left_pile.options(urwid.WEIGHT, 1)
        self.delegate.parent.left_pile.contents[0] = (overlay, options)

    def timestamp(self):
        return self.timestamp

class TabButton(urwid.Button):
    button_left = urwid.Text("[")
    button_right = urwid.Text("]")

class AnnounceStream(urwid.WidgetWrap):
    def __init__(self, app, parent):
        self.app = app
        self.parent = parent
        self.started = False
        self.timeout = self.app.config["textui"]["animation_interval"]*2
        self.ilb = None
        self.no_content = True
        self.current_tab = "nodes"
        self.show_destination = False
        self.search_text = ""

        self.added_entries = []
        self.widget_list = []

        self.tab_nodes = TabButton("Nodes (0)", on_press=self.show_nodes_tab)
        self.tab_peers = TabButton("Peers (0)", on_press=self.show_peers_tab)
        self.tab_pn = TabButton("Propagation Nodes (0)", on_press=self.show_pn_tab)

        self.tab_bar = urwid.Columns([
            ('weight', 1, self.tab_nodes),
            ('weight', 1, self.tab_peers),
            ('weight', 3, self.tab_pn),
        ], dividechars=1)

        self.search_edit = urwid.Edit(caption="Search: ")
        urwid.connect_signal(self.search_edit, 'change', self.on_search_change)

        self.display_toggle = TabButton("Show: Name", on_press=self.toggle_display_mode)

        self.filter_bar = urwid.Columns([
            ('weight', 2, self.search_edit),
            ('weight', 1, self.display_toggle),
        ], dividechars=1)

        self.update_widget_list()

        self.ilb = ExceptionHandlingListBox(
            self.widget_list,
            on_selection_change=self.list_selection,
            initialization_is_selection_change=False,
            #modifier_key=MODIFIER_KEY.CTRL,
            #highlight_offFocus="list_off_focus"
        )

        self.pile = urwid.Pile([
            ('pack', self.tab_bar),
            ('pack', self.filter_bar),
            ('weight', 1, self.ilb),
        ])

        self.display_widget = self.pile
        super().__init__(urwid.LineBox(self.display_widget, title="Announce Stream"))

    def keypress(self, size, key):
        if key == "up" and self.pile.focus == self.tab_bar:
            nomadnet.NomadNetworkApp.get_shared_instance().ui.main_display.frame.focus_position = "header"
        elif key == "ctrl x":
            self.delete_selected_entry()

        return super(AnnounceStream, self).keypress(size, key)

    def on_search_change(self, widget, text):
        self.search_text = text.lower()
        self.update_widget_list()

    def toggle_display_mode(self, button):
        self.show_destination = not self.show_destination
        if self.show_destination:
            self.display_toggle.set_label("Show: Dest")
        else:
            self.display_toggle.set_label("Show: Name")
        self.update_widget_list()

    def delete_selected_entry(self):
        sel = self.ilb.get_selected_item()
        if sel != None and hasattr(sel, "original_widget") and sel.original_widget:
            if hasattr(sel.original_widget, "timestamp"):
                self.app.directory.remove_announce_with_timestamp(sel.original_widget.timestamp)
                self.rebuild_widget_list()

    def rebuild_widget_list(self):
        self.no_content = True
        self.added_entries = []
        self.widget_list = []
        self.update_widget_list()

    def update_widget_list(self):
        self.widget_list = []
        new_entries = []

        node_count = 0
        peer_count = 0
        pn_count = 0

        for e in self.app.directory.announce_stream:
            announce_type = e[3]

            if self.search_text:
                try:
                    announce_data = e[2].decode("utf-8").lower()
                except:
                    announce_data = ""
                if self.search_text not in announce_data:
                    continue

            if announce_type == "node" or announce_type == True:
                node_count += 1
                if self.current_tab == "nodes": new_entries.append(e)
            
            elif announce_type == "peer" or announce_type == False:
                peer_count += 1
                if self.current_tab == "peers": new_entries.append(e)

            elif announce_type == "pn":
                pn_count += 1
                if self.current_tab == "pn": new_entries.append(e)

        for e in new_entries:
            nw = AnnounceStreamEntry(self.app, e, self, show_destination=self.show_destination)
            nw.timestamp = e[0]
            self.widget_list.append(nw)

        if len(new_entries) > 0:
            self.no_content = False
        else:
            self.no_content = True
            self.widget_list = [urwid.Text(f"No {self.current_tab} announces", align='center')]

        self.tab_nodes.set_label(f"Nodes ({node_count})")
        self.tab_peers.set_label(f"Peers ({peer_count})")
        self.tab_pn.set_label(f"Propagation Nodes ({pn_count})")

        if self.ilb:
            self.ilb.set_body(self.widget_list)

    def show_nodes_tab(self, button):
        self.current_tab = "nodes"
        self.update_widget_list()

    def show_peers_tab(self, button):
        self.current_tab = "peers"
        self.update_widget_list()

    def show_pn_tab(self, button):
        self.current_tab = "pn"
        self.update_widget_list()

    def list_selection(self, arg1, arg2):
        pass

    def update(self):
        self.update_widget_list()

    def update_callback(self, loop=None, user_data=None):
        self.update()
        if self.started:
            self.app.ui.loop.set_alarm_in(self.timeout, self.update_callback)

    def start(self):
        was_started = self.started
        self.started = True
        if not was_started:
            self.update_callback()

    def stop(self):
        self.started = False

class SelectText(urwid.Text):
    _selectable = True

    signals = ["click"]

    def keypress(self, size, key):
        """
        Send 'click' signal on 'activate' command.
        """
        if self._command_map[key] != urwid.ACTIVATE:
            return key

        self._emit('click')

    def mouse_event(self, size, event, button, x, y, focus):
        """
        Send 'click' signal on button 1 press.
        """
        if button != 1 or not urwid.util.is_mouse_press(event):
            return False

        self._emit('click')
        return True

class ListDialogLineBox(urwid.LineBox):
    def keypress(self, size, key):
        if key == "esc":
            self.delegate.close_list_dialogs()
        else:
            return super(ListDialogLineBox, self).keypress(size, key)

class KnownNodeInfo(urwid.WidgetWrap):
    def keypress(self, size, key):
        if key == "esc":
            options = self.parent.left_pile.options(height_type=urwid.WEIGHT, height_amount=1)
            self.parent.left_pile.contents[0] = (self.parent.known_nodes_display, options)
        else:
            return super(KnownNodeInfo, self).keypress(size, key)

    def __init__(self, node_hash):
        self.app = nomadnet.NomadNetworkApp.get_shared_instance()
        self.parent = self.app.ui.main_display.sub_displays.network_display
        self.pn_changed = False
        g = self.app.ui.glyphs

        source_hash  = node_hash
        node_ident   = RNS.Identity.recall(node_hash)
        time_format  = self.app.time_format
        trust_level  = self.app.directory.trust_level(source_hash)
        trust_str    = ""
        node_entry   = self.app.directory.find(source_hash)
        sort_str     = self.app.directory.sort_rank(source_hash)
        if sort_str == None:
            sort_str = "None"
        else:
            sort_str = str(sort_str)

        if node_entry == None:
            display_str = self.app.directory.simplest_display_str(source_hash)
        else:
            display_str = strip_modifiers(node_entry.display_name)

        addr_str     = "<"+RNS.hexrep(source_hash, delimit=False)+">"

        if display_str == None:
            display_str = addr_str

        pn_hash = RNS.Destination.hash_from_name_and_identity("lxmf.propagation", node_ident)

        if node_ident != None:
            lxmf_addr_str = g["sent"]+" LXMF Propagation Node Address is "+RNS.prettyhexrep(pn_hash)
        else:
            lxmf_addr_str = "No associated Propagation Node known"


        type_string = "Nomad Network Node " + g["node"]

        if trust_level == DirectoryEntry.UNTRUSTED:
            trust_str     = "Untrusted"
            symbol        = g["cross"]
            style         = "list_untrusted"
        elif trust_level == DirectoryEntry.UNKNOWN:
            trust_str     = "Unknown"
            symbol        = g["unknown"]
            style         = "list_unknown"
        elif trust_level == DirectoryEntry.TRUSTED:
            trust_str     = "Trusted"
            symbol        = g["check"]
            style         = "list_trusted"
        elif trust_level == DirectoryEntry.WARNING:
            trust_str     = "Warning"
            symbol        = g["warning"]
            style         = "list_warning"
        else:
            trust_str     = "Warning"
            symbol        = g["warning"]
            style         = "list_untrusted"

        if trust_level == DirectoryEntry.UNTRUSTED:
            untrusted_selected = True
            unknown_selected   = False
            trusted_selected   = False
        elif trust_level == DirectoryEntry.UNKNOWN:
            untrusted_selected = False
            unknown_selected   = True
            trusted_selected   = False
        elif trust_level == DirectoryEntry.TRUSTED:
            untrusted_selected = False
            unknown_selected   = False
            trusted_selected   = True

        trust_button_group = []
        r_untrusted = urwid.RadioButton(trust_button_group, "Untrusted", state=untrusted_selected)
        r_unknown   = urwid.RadioButton(trust_button_group, "Unknown", state=unknown_selected)
        r_trusted   = urwid.RadioButton(trust_button_group, "Trusted", state=trusted_selected)

        e_name = urwid.Edit(caption="Name      : ",edit_text=display_str)
        e_sort = urwid.Edit(caption="Sort Rank : ",edit_text=sort_str)

        node_ident = RNS.Identity.recall(source_hash)
        op_hash = None
        op_str = None
        if node_ident != None:
            op_hash = RNS.Destination.hash_from_name_and_identity("lxmf.delivery", node_ident)
            op_str = self.app.directory.simplest_display_str(op_hash)
        else:
            op_str = "Unknown"

        def show_known_nodes(sender):
            options = self.parent.left_pile.options(height_type=urwid.WEIGHT, height_amount=1)
            self.parent.left_pile.contents[0] = (self.parent.known_nodes_display, options)

        def connect(sender):
            self.parent.browser.retrieve_url(RNS.hexrep(source_hash, delimit=False))
            show_known_nodes(None)

        def msg_op(sender):
            show_known_nodes(None)
            if node_ident != None:
                try:
                    existing_conversations = nomadnet.Conversation.conversation_list(self.app)

                    source_hash_text = RNS.hexrep(op_hash, delimit=False)
                    display_name = op_str

                    if not source_hash_text in [c[0] for c in existing_conversations]:
                        entry = DirectoryEntry(source_hash, display_name, trust_level)
                        self.app.directory.remember(entry)

                        new_conversation = nomadnet.Conversation(source_hash_text, nomadnet.NomadNetworkApp.get_shared_instance(), initiator=True)
                        self.app.ui.main_display.sub_displays.conversations_display.update_conversation_list()

                    self.app.ui.main_display.sub_displays.conversations_display.display_conversation(None, source_hash_text)
                    self.app.ui.main_display.show_conversations(None)

                except Exception as e:
                    RNS.log("Error while starting conversation from node info. The contained exception was: "+str(e), RNS.LOG_ERROR)

        def pn_change(sender, userdata):
            self.pn_changed = True

        def ident_change(sender, userdata):
            pass

        propagation_node_checkbox = urwid.CheckBox("Use as default propagation node", state=(self.app.get_user_selected_propagation_node() == pn_hash), on_state_change=pn_change)
        connect_identify_checkbox = urwid.CheckBox("Identify when connecting", state=self.app.directory.should_identify_on_connect(source_hash), on_state_change=ident_change)

        def save_node(sender):
            if self.pn_changed:
                if propagation_node_checkbox.get_state():
                    self.app.set_user_selected_propagation_node(pn_hash)
                else:
                    self.app.set_user_selected_propagation_node(None)

            trust_level = DirectoryEntry.UNTRUSTED
            if r_unknown.get_state() == True:
                trust_level = DirectoryEntry.UNKNOWN

            if r_trusted.get_state() == True:
                trust_level = DirectoryEntry.TRUSTED

            display_str = e_name.get_edit_text()
            sort_rank = e_sort.get_edit_text()
            try:
                if int(sort_rank) >= 0:
                    sort_rank = int(sort_rank)
                else:
                    sort_rank = None
            except:
                sort_rank = None

            node_entry = DirectoryEntry(source_hash, display_name=display_str, trust_level=trust_level, hosts_node=True, identify_on_connect=connect_identify_checkbox.get_state(), sort_rank=sort_rank)
            self.app.directory.remember(node_entry)
            self.app.ui.main_display.sub_displays.network_display.directory_change_callback()

            if trust_level == DirectoryEntry.TRUSTED:
                self.app.autoselect_propagation_node()

            show_known_nodes(None)

        back_button = (urwid.WEIGHT, 0.2, urwid.Button("Back", on_press=show_known_nodes))
        connect_button = (urwid.WEIGHT, 0.2, urwid.Button("Connect", on_press=connect))
        save_button = (urwid.WEIGHT, 0.2, urwid.Button("Save", on_press=save_node))
        msg_button = (urwid.WEIGHT, 0.2, urwid.Button("Msg Op", on_press=msg_op))
        bdiv = (urwid.WEIGHT, 0.02, urwid.Text(""))

        button_columns = urwid.Columns([back_button, bdiv, connect_button, bdiv, msg_button, bdiv, save_button])

        pile_widgets = [
            urwid.Text("Type      : "+type_string, align=urwid.LEFT),
            e_name,
            urwid.Text("Node Addr : "+addr_str, align=urwid.LEFT),
            e_sort,
            urwid.Divider(g["divider1"]),
            urwid.Text(lxmf_addr_str, align=urwid.CENTER),
            urwid.Divider(g["divider1"]),
            propagation_node_checkbox,
            connect_identify_checkbox,
            urwid.Divider(g["divider1"]),
            r_untrusted,
            r_unknown,
            r_trusted,
            urwid.Divider(g["divider1"]),
            button_columns
        ]

        operator_entry = urwid.Text("Operator  : "+op_str, align=urwid.LEFT)
        pile_widgets.insert(3, operator_entry)

        hops = RNS.Transport.hops_to(source_hash)
        if hops == 1:
            str_s = ""
        else:
            str_s = "s"

        if hops != RNS.Transport.PATHFINDER_M:
            hops_str = str(hops)+" hop"+str_s
        else:
            hops_str = "Unknown"

        operator_entry = urwid.Text("Distance  : "+hops_str, align=urwid.LEFT)
        pile_widgets.insert(4, operator_entry)

        pile = urwid.Pile(pile_widgets)

        pile.focus_position = len(pile.contents)-1
        button_columns.focus_position = 0


        self.display_widget = urwid.Filler(pile, valign=urwid.TOP, height=urwid.PACK)

        super().__init__(urwid.LineBox(self.display_widget, title="Node Info"))


# Yes, this is weird. There is a bug in Urwid/ILB that causes
# an indexing exception when the list is very small vertically.
# This mitigates it.
class ExceptionHandlingListBox(IndicativeListBox):
    def keypress(self, size, key):
        try:
            return super(ExceptionHandlingListBox, self).keypress(size, key)

        except Exception as e:
            if key == "up":
                nomadnet.NomadNetworkApp.get_shared_instance().ui.main_display.frame.focus_position = "header"
            elif key == "down":
                nomadnet.NomadNetworkApp.get_shared_instance().ui.main_display.sub_displays.network_display.left_pile.focus_position = 1
            else:
                RNS.log("An error occurred while processing an interface event. The contained exception was: "+str(e), RNS.LOG_ERROR)


class KnownNodes(urwid.WidgetWrap):
    def __init__(self, app):
        self.app = app
        self.node_list = app.directory.known_nodes()
        g = self.app.ui.glyphs

        self.widget_list = self.make_node_widgets()

        self.ilb = ExceptionHandlingListBox(
            self.widget_list,
            on_selection_change=self.node_list_selection,
            initialization_is_selection_change=False,
            highlight_offFocus="list_off_focus"
        )

        if len(self.node_list) > 0:
            self.display_widget = self.ilb
            widget_style = None
            self.no_content = False
        else:
            self.no_content = True
            widget_style = "inactive_text"
            self.pile = urwid.Pile([
                urwid.Text(("warning_text", g["info"]+"\n"), align=urwid.CENTER),
                SelectText(
                    (
                        "warning_text",
                        "Currently, no nodes are saved\n\nCtrl+L to view the announce stream\n\n",
                    ),
                    align=urwid.CENTER,
                ),
            ])
            self.display_widget = urwid.Filler(self.pile, valign=urwid.TOP, height=urwid.PACK)

        super().__init__(urwid.AttrMap(urwid.LineBox(self.display_widget, title="Saved Nodes"), widget_style))

    def keypress(self, size, key):
        if key == "up" and (self.no_content or self.ilb.first_item_is_selected()):
            nomadnet.NomadNetworkApp.get_shared_instance().ui.main_display.frame.focus_position = "header"
        elif key == "ctrl x":
            self.delete_selected_entry()

        return super(KnownNodes, self).keypress(size, key)


    def node_list_selection(self, arg1, arg2):
        pass

    def connect_node(self, event, node):
        source_hash = node.source_hash
        trust_level = node.trust_level
        trust_level  = self.app.directory.trust_level(source_hash)
        display_str = self.app.directory.simplest_display_str(source_hash)

        parent = self.app.ui.main_display.sub_displays.network_display

        def dismiss_dialog(sender):
            self.delegate.close_list_dialogs()

        def confirmed(sender):
            self.delegate.browser.retrieve_url(RNS.hexrep(source_hash, delimit=False))
            self.delegate.close_list_dialogs()

        def show_info(sender):
            info_widget = KnownNodeInfo(source_hash)
            options = parent.left_pile.options(height_type=urwid.WEIGHT, height_amount=1)
            parent.left_pile.contents[0] = (info_widget, options)


        dialog = ListDialogLineBox(
            urwid.Pile([
                urwid.Text("Connect to node\n"+self.app.directory.simplest_display_str(source_hash)+"\n", align=urwid.CENTER),
                urwid.Columns([
                    (urwid.WEIGHT, 0.45, urwid.Button("Yes", on_press=confirmed)),
                    (urwid.WEIGHT, 0.1, urwid.Text("")),
                    (urwid.WEIGHT, 0.45, urwid.Button("No", on_press=dismiss_dialog)),
                    (urwid.WEIGHT, 0.1, urwid.Text("")),
                    (urwid.WEIGHT, 0.45, urwid.Button("Info", on_press=show_info))])
            ]), title="?"
        )
        dialog.delegate = self.delegate
        bottom = self

        overlay = urwid.Overlay(dialog, bottom, align=urwid.CENTER, width=urwid.RELATIVE_100, valign=urwid.MIDDLE, height=urwid.PACK, left=2, right=2)

        options = self.delegate.left_pile.options(urwid.WEIGHT, 1)
        self.delegate.left_pile.contents[0] = (overlay, options)

    def delete_selected_entry(self):
        si = self.ilb.get_selected_item()
        if si != None:
            source_hash = si.original_widget.source_hash

            def dismiss_dialog(sender):
                self.delegate.close_list_dialogs()

            def confirmed(sender):
                self.app.directory.forget(source_hash)
                self.rebuild_widget_list()
                self.delegate.close_list_dialogs()


            dialog = ListDialogLineBox(
                urwid.Pile([
                    urwid.Text("Delete Node\n"+self.app.directory.simplest_display_str(source_hash)+"\n", align=urwid.CENTER),
                    urwid.Columns([
                        (urwid.WEIGHT, 0.45, urwid.Button("Yes", on_press=confirmed)),
                        (urwid.WEIGHT, 0.1, urwid.Text("")),
                        (urwid.WEIGHT, 0.45, urwid.Button("No", on_press=dismiss_dialog)),
                    ])
                ]), title="?"
            )
            dialog.delegate = self.delegate
            bottom = self

            overlay = urwid.Overlay(
                dialog,
                bottom,
                align=urwid.CENTER,
                width=urwid.RELATIVE_100,
                valign=urwid.MIDDLE,
                height=urwid.PACK,
                left=2,
                right=2,
            )

            options = self.delegate.left_pile.options(urwid.WEIGHT, 1)
            self.delegate.left_pile.contents[0] = (overlay, options)


    def rebuild_widget_list(self):
        self.node_list = self.app.directory.known_nodes()
        self.widget_list = self.make_node_widgets()
        self.ilb.set_body(self.widget_list)
        if len(self.widget_list) > 0:
            self.no_content = False
        else:
            self.no_content = True
            self.delegate.reinit_known_nodes()

    def make_node_widgets(self):
        widget_list = []
        for node_entry in self.node_list:
            # TODO: Implement this
            ne = NodeEntry(self.app, node_entry, self)
            ne.source_hash = node_entry.source_hash
            widget_list.append(ne)

        # TODO: Sort list
        return widget_list

class NodeEntry(urwid.WidgetWrap):
    def __init__(self, app, node, delegate):
        source_hash = node.source_hash
        trust_level = node.trust_level

        self.app = app
        g = self.app.ui.glyphs

        trust_level  = self.app.directory.trust_level(source_hash)
        display_str =  self.app.directory.simplest_display_str(source_hash, san=False)

        if trust_level == DirectoryEntry.UNTRUSTED:
            symbol        = g["cross"]
            style         = "list_untrusted"
            focus_style   = "list_focus_untrusted"
        elif trust_level == DirectoryEntry.UNKNOWN:
            symbol        = g["unknown"]
            style         = "list_unknown"
            focus_style   = "list_focus"
        elif trust_level == DirectoryEntry.TRUSTED:
            symbol        = g["check"]
            style         = "list_trusted"
            focus_style   = "list_focus_trusted"
        elif trust_level == DirectoryEntry.WARNING:
            symbol        = g["warning"]
            style         = "list_warning"
            focus_style   = "list_focus"
        else:
            symbol        = g["warning"]
            style         = "list_untrusted"
            focus_style   = "list_focus_untrusted"

        type_symbol = g["node"]

        widget = ListEntry(type_symbol+" "+display_str)
        urwid.connect_signal(widget, "click", delegate.connect_node, node)

        self.display_widget = urwid.AttrMap(widget, style, focus_style)
        self.display_widget.source_hash = source_hash
        super().__init__(self.display_widget)


class AnnounceTime(urwid.WidgetWrap):
    def __init__(self, app):
        self.started = False
        self.app = app
        self.timeout = self.app.config["textui"]["animation_interval"]
        self.display_widget = urwid.Text("")
        self.update_time()

        super().__init__(self.display_widget)

    def update_time(self):
        self.last_announce_string = "Never"
        if self.app.peer_settings["last_announce"] != None:
            self.last_announce_string = pretty_date(int(self.app.peer_settings["last_announce"]))

        self.display_widget.set_text("Announced : "+self.last_announce_string)

    def update_time_callback(self, loop=None, user_data=None):
        self.update_time()
        if self.started:
            self.app.ui.loop.set_alarm_in(self.timeout, self.update_time_callback)

    def start(self):
        was_started = self.started
        self.started = True
        if not was_started:
            self.update_time_callback()

    def stop(self):
        self.started = False


class NodeAnnounceTime(urwid.WidgetWrap):
    def __init__(self, app):
        self.started = False
        self.app = app
        self.timeout = self.app.config["textui"]["animation_interval"]
        self.display_widget = urwid.Text("")
        self.update_time()

        super().__init__(self.display_widget)

    def update_time(self):
        self.last_announce_string = "Never"
        if self.app.peer_settings["node_last_announce"] != None:
            self.last_announce_string = pretty_date(int(self.app.peer_settings["node_last_announce"]))

        self.display_widget.set_text("Last Announce  : "+self.last_announce_string)

    def update_time_callback(self, loop=None, user_data=None):
        self.update_time()
        if self.started:
            self.app.ui.loop.set_alarm_in(self.timeout, self.update_time_callback)

    def start(self):
        was_started = self.started
        self.started = True
        if not was_started:
            self.update_time_callback()

    def stop(self):
        self.started = False

class NodeActiveConnections(urwid.WidgetWrap):
    def __init__(self, app):
        self.started = False
        self.app = app
        self.timeout = self.app.config["textui"]["animation_interval"]
        self.display_widget = urwid.Text("")
        self.update_stat()

        super().__init__(self.display_widget)

    def update_stat(self):
        self.stat_string = "None"
        if self.app.node != None:
            self.stat_string = str(len(self.app.node.destination.links))

        self.display_widget.set_text("Connected Now  : "+self.stat_string)

    def update_stat_callback(self, loop=None, user_data=None):
        self.update_stat()
        if self.started:
            self.app.ui.loop.set_alarm_in(self.timeout, self.update_stat_callback)

    def start(self):
        was_started = self.started
        self.started = True
        if not was_started:
            self.update_stat_callback()

    def stop(self):
        self.started = False

class NodeStorageStats(urwid.WidgetWrap):
    def __init__(self, app):
        self.started = False
        self.app = app
        self.timeout = self.app.config["textui"]["animation_interval"]
        self.display_widget = urwid.Text("")
        self.update_stat()

        super().__init__(self.display_widget)

    def update_stat(self):
        self.stat_string = "None"
        if self.app.node != None and not self.app.disable_propagation:

            limit = self.app.message_router.message_storage_limit
            used = self.app.message_router.message_storage_size()

            if limit != None and used != None:
                pct = round((used/limit)*100, 1)
                pct_str = str(pct)+"%, "
                limit_str = " of "+RNS.prettysize(limit)
            else:
                limit_str = ""
                pct_str = ""

            self.stat_string = pct_str+RNS.prettysize(used)+limit_str

        self.display_widget.set_text("LXMF Storage   : "+self.stat_string)

    def update_stat_callback(self, loop=None, user_data=None):
        self.update_stat()
        if self.started:
            self.app.ui.loop.set_alarm_in(self.timeout, self.update_stat_callback)

    def start(self):
        was_started = self.started
        self.started = True
        if not was_started:
            self.update_stat_callback()

    def stop(self):
        self.started = False

class NodeTotalConnections(urwid.WidgetWrap):
    def __init__(self, app):
        self.started = False
        self.app = app
        self.timeout = self.app.config["textui"]["animation_interval"]
        self.display_widget = urwid.Text("")
        self.update_stat()

        super().__init__(self.display_widget)

    def update_stat(self):
        self.stat_string = "None"
        if self.app.node != None:
            self.stat_string = str(self.app.peer_settings["node_connects"])

        self.display_widget.set_text("Total Connects : "+self.stat_string)

    def update_stat_callback(self, loop=None, user_data=None):
        self.update_stat()
        if self.started:
            self.app.ui.loop.set_alarm_in(self.timeout, self.update_stat_callback)

    def start(self):
        was_started = self.started
        self.started = True
        if not was_started:
            self.update_stat_callback()

    def stop(self):
        self.started = False


class NodeTotalPages(urwid.WidgetWrap):
    def __init__(self, app):
        self.started = False
        self.app = app
        self.timeout = self.app.config["textui"]["animation_interval"]
        self.display_widget = urwid.Text("")
        self.update_stat()

        super().__init__(self.display_widget)

    def update_stat(self):
        self.stat_string = "None"
        if self.app.node != None:
            self.stat_string = str(self.app.peer_settings["served_page_requests"])

        self.display_widget.set_text("Served Pages   : "+self.stat_string)

    def update_stat_callback(self, loop=None, user_data=None):
        self.update_stat()
        if self.started:
            self.app.ui.loop.set_alarm_in(self.timeout, self.update_stat_callback)

    def start(self):
        was_started = self.started
        self.started = True
        if not was_started:
            self.update_stat_callback()

    def stop(self):
        self.started = False


class NodeTotalFiles(urwid.WidgetWrap):
    def __init__(self, app):
        self.started = False
        self.app = app
        self.timeout = self.app.config["textui"]["animation_interval"]
        self.display_widget = urwid.Text("")
        self.update_stat()

        super().__init__(self.display_widget)

    def update_stat(self):
        self.stat_string = "None"
        if self.app.node != None:
            self.stat_string = str(self.app.peer_settings["served_file_requests"])

        self.display_widget.set_text("Served Files   : "+self.stat_string)

    def update_stat_callback(self, loop=None, user_data=None):
        self.update_stat()
        if self.started:
            self.app.ui.loop.set_alarm_in(self.timeout, self.update_stat_callback)

    def start(self):
        was_started = self.started
        self.started = True
        if not was_started:
            self.update_stat_callback()

    def stop(self):
        self.started = False


class LocalPeer(urwid.WidgetWrap):
    announce_timer = None

    def __init__(self, app, parent):
        self.app = app
        self.parent = parent
        g = self.app.ui.glyphs
        self.dialog_open = False
        display_name = self.app.lxmf_destination.display_name
        if display_name == None:
            display_name = ""

        t_id =           urwid.Text("LXMF Addr : "+RNS.prettyhexrep(self.app.lxmf_destination.hash))
        i_id =           urwid.Text("Identity  : "+RNS.prettyhexrep(self.app.identity.hash))
        e_name = urwid.Edit(caption="Name      : ", edit_text=display_name)

        def save_query(sender):
            def dismiss_dialog(sender):
                self.dialog_open = False
                self.parent.left_pile.contents[1] = (LocalPeer(self.app, self.parent), options)

            self.app.set_display_name(e_name.get_edit_text())

            dialog = DialogLineBox(
                urwid.Pile([
                    urwid.Text("\n\n\nSaved\n\n", align=urwid.CENTER),
                    urwid.Button("OK", on_press=dismiss_dialog)
                ]), title=g["info"]
            )
            dialog.delegate = self
            bottom = self

            #overlay = urwid.Overlay(dialog, bottom, align=urwid.CENTER, width=urwid.RELATIVE_100, valign=urwid.MIDDLE, height=urwid.PACK, left=4, right=4)
            overlay = dialog
            options = self.parent.left_pile.options(height_type=urwid.PACK, height_amount=None)
            self.dialog_open = True
            self.parent.left_pile.contents[1] = (overlay, options)

        def announce_query(sender):
            def dismiss_dialog(sender):
                self.dialog_open = False
                options = self.parent.left_pile.options(height_type=urwid.PACK, height_amount=None)
                self.parent.left_pile.contents[1] = (LocalPeer(self.app, self.parent), options)

            self.app.announce_now()

            dialog = DialogLineBox(
                urwid.Pile([
                    urwid.Text("\n\n\nAnnounce Sent\n\n\n", align=urwid.CENTER),
                    urwid.Button("OK", on_press=dismiss_dialog)
                ]), title=g["info"]
            )
            dialog.delegate = self
            bottom = self

            #overlay = urwid.Overlay(dialog, bottom, align=urwid.CENTER, width=urwid.RELATIVE_100, valign=urwid.MIDDLE, height=urwid.PACK, left=4, right=4)
            overlay = dialog

            self.dialog_open = True
            options = self.parent.left_pile.options(height_type=urwid.PACK, height_amount=None)
            self.parent.left_pile.contents[1] = (overlay, options)

        def node_info_query(sender):
            options = self.parent.left_pile.options(height_type=urwid.PACK, height_amount=None)
            self.parent.left_pile.contents[1] = (self.parent.node_info_display, options)

        if LocalPeer.announce_timer == None:
            self.t_last_announce = AnnounceTime(self.app)
            LocalPeer.announce_timer = self.t_last_announce
        else:
            self.t_last_announce = LocalPeer.announce_timer
            self.t_last_announce.update_time()

        announce_button = urwid.Button("Announce Now", on_press=announce_query)

        self.display_widget = urwid.Pile(
            [
                t_id,
                i_id,
                e_name,
                urwid.Divider(g["divider1"]),
                self.t_last_announce,
                announce_button,
                urwid.Divider(g["divider1"]),
                urwid.Columns([
                    (urwid.WEIGHT, 0.45, urwid.Button("Save", on_press=save_query)),
                    (urwid.WEIGHT, 0.1, urwid.Text("")),
                    (urwid.WEIGHT, 0.45, urwid.Button("Node Info", on_press=node_info_query)),
                ]),
            ]
        )

        super().__init__(urwid.LineBox(self.display_widget, title="Local Peer Info"))

    def start(self):
        self.t_last_announce.start()


class NodeInfo(urwid.WidgetWrap):
    announce_timer = None
    links_timer = None
    conns_timer = None
    pages_timer = None
    files_timer = None
    storage_timer = None

    def __init__(self, app, parent):
        self.app = app
        self.parent = parent
        g = self.app.ui.glyphs

        self.dialog_open = False

        widget_style = ""

        def show_peer_info(sender):
            options = self.parent.left_pile.options(height_type=urwid.PACK, height_amount=None)
            self.parent.left_pile.contents[1] = (LocalPeer(self.app, self.parent), options)

        if self.app.enable_node:
            if self.app.node != None:
                display_name = self.app.node.name
            else:
                display_name = None

            if display_name == None:
                display_name = ""

            t_id = urwid.Text("Addr : "+RNS.hexrep(self.app.node.destination.hash, delimit=False))
            e_name = urwid.Text("Name : "+display_name)

            def stats_query(sender):
                self.app.peer_settings["node_connects"] = 0
                self.app.peer_settings["served_page_requests"] = 0
                self.app.peer_settings["served_file_requests"] = 0
                self.app.save_peer_settings()

            def announce_query(sender):
                def dismiss_dialog(sender):
                    self.dialog_open = False
                    options = self.parent.left_pile.options(height_type=urwid.PACK, height_amount=None)
                    self.parent.left_pile.contents[1] = (NodeInfo(self.app, self.parent), options)

                self.app.node.announce()

                dialog = DialogLineBox(
                    urwid.Pile([
                        urwid.Text("\n\n\nAnnounce Sent\n\n", align=urwid.CENTER),
                        urwid.Button("OK", on_press=dismiss_dialog)
                    ]), title=g["info"]
                )
                dialog.delegate = self
                bottom = self

                #overlay = urwid.Overlay(dialog, bottom, align=urwid.CENTER, width=urwid.RELATIVE_100, valign=urwid.MIDDLE, height=urwid.PACK, left=4, right=4)
                overlay = dialog

                self.dialog_open = True
                options = self.parent.left_pile.options(height_type=urwid.PACK, height_amount=None)
                self.parent.left_pile.contents[1] = (overlay, options)

            def connect_query(sender):
                self.parent.browser.retrieve_url(RNS.hexrep(self.app.node.destination.hash, delimit=False))

            if NodeInfo.announce_timer == None:
                self.t_last_announce = NodeAnnounceTime(self.app)
                NodeInfo.announce_timer = self.t_last_announce
            else:
                self.t_last_announce = NodeInfo.announce_timer
                self.t_last_announce.update_time()

            if NodeInfo.links_timer == None:
                self.t_active_links = NodeActiveConnections(self.app)
                NodeInfo.links_timer = self.t_active_links
            else:
                self.t_active_links = NodeInfo.links_timer
                self.t_active_links.update_stat()

            if NodeInfo.storage_timer == None:
                self.t_storage_stats = NodeStorageStats(self.app)
                NodeInfo.storage_timer = self.t_storage_stats
            else:
                self.t_storage_stats = NodeInfo.storage_timer
                self.t_storage_stats.update_stat()

            if NodeInfo.conns_timer == None:
                self.t_total_connections = NodeTotalConnections(self.app)
                NodeInfo.conns_timer = self.t_total_connections
            else:
                self.t_total_connections = NodeInfo.conns_timer
                self.t_total_connections.update_stat()

            if NodeInfo.pages_timer == None:
                self.t_total_pages = NodeTotalPages(self.app)
                NodeInfo.pages_timer = self.t_total_pages
            else:
                self.t_total_pages = NodeInfo.pages_timer
                self.t_total_pages.update_stat()

            if NodeInfo.files_timer == None:
                self.t_total_files = NodeTotalFiles(self.app)
                NodeInfo.files_timer = self.t_total_files
            else:
                self.t_total_files = NodeInfo.files_timer
                self.t_total_files.update_stat()

            lxmf_addr_str = g["sent"]+" LXMF Propagation Node Address is "+RNS.prettyhexrep(RNS.Destination.hash_from_name_and_identity("lxmf.propagation", self.app.node.destination.identity))
            e_lxmf = urwid.Text(lxmf_addr_str, align=urwid.CENTER)

            announce_button = urwid.Button("Announce", on_press=announce_query)
            connect_button = urwid.Button("Browse", on_press=connect_query)
            reset_button = urwid.Button("Rst Stats", on_press=stats_query)

            if not self.app.disable_propagation:
                pile = urwid.Pile([
                    t_id,
                    e_name,
                    urwid.Divider(g["divider1"]),
                    e_lxmf,
                    urwid.Divider(g["divider1"]),
                    self.t_last_announce,
                    self.t_storage_stats,
                    self.t_active_links,
                    self.t_total_connections,
                    self.t_total_pages,
                    self.t_total_files,
                    urwid.Divider(g["divider1"]),
                    urwid.Columns([
                        (urwid.WEIGHT, 5, urwid.Button("Back", on_press=show_peer_info)),
                        (urwid.WEIGHT, 0.5, urwid.Text("")),
                        (urwid.WEIGHT, 6, connect_button),
                        (urwid.WEIGHT, 0.5, urwid.Text("")),
                        (urwid.WEIGHT, 8, reset_button),
                        (urwid.WEIGHT, 0.5, urwid.Text("")),
                        (urwid.WEIGHT, 7, announce_button),
                    ])
                ])
            else:
                pile = urwid.Pile([
                t_id,
                e_name,
                urwid.Divider(g["divider1"]),
                self.t_last_announce,
                self.t_storage_stats,
                self.t_active_links,
                self.t_total_connections,
                self.t_total_pages,
                self.t_total_files,
                urwid.Divider(g["divider1"]),
                urwid.Columns([
                    (urwid.WEIGHT, 5, urwid.Button("Back", on_press=show_peer_info)),
                    (urwid.WEIGHT, 0.5, urwid.Text("")),
                    (urwid.WEIGHT, 6, connect_button),
                    (urwid.WEIGHT, 0.5, urwid.Text("")),
                    (urwid.WEIGHT, 8, reset_button),
                    (urwid.WEIGHT, 0.5, urwid.Text("")),
                    (urwid.WEIGHT, 7, announce_button),
                ])
            ])
        else:
            pile = urwid.Pile([
                urwid.Text("\n"+g["info"], align=urwid.CENTER),
                urwid.Text("\nThis instance is not hosting a node\n\n", align=urwid.CENTER),
                urwid.Padding(urwid.Button("Back", on_press=show_peer_info), urwid.CENTER, urwid.PACK)
            ])

        self.display_widget = pile

        super().__init__(urwid.AttrMap(urwid.LineBox(self.display_widget, title="Local Node Info"), widget_style))

    def start(self):
        if self.app.node != None:
            self.t_last_announce.start()
            self.t_active_links.start()
            self.t_total_connections.start()
            self.t_total_pages.start()
            self.t_total_files.start()


class UpdatingText(urwid.WidgetWrap):
    def __init__(self, app, title, value_method, append_text=""):
        self.started = False
        self.app = app
        self.timeout = self.app.config["textui"]["animation_interval"]*5
        self.display_widget = urwid.Text("")
        self.value = None
        self.value_method = value_method
        self.title = title
        self.append_text = append_text
        self.update()

        super().__init__(self.display_widget)

    def update(self):
        self.value = self.value_method()
        self.display_widget.set_text(self.title+str(self.value)+str(self.append_text))

    def update_callback(self, loop=None, user_data=None):
        self.update()
        if self.started:
            self.app.ui.loop.set_alarm_in(self.timeout, self.update_callback)

    def start(self):
        was_started = self.started
        self.started = True
        if not was_started:
            self.update_callback()

    def stop(self):
        self.started = False

class NetworkStats(urwid.WidgetWrap):
    def __init__(self, app, parent):
        self.app = app
        self.parent = parent

        def get_num_peers():
            return self.app.directory.number_of_known_peers(lookback_seconds=30*60)


        def get_num_nodes():
            return self.app.directory.number_of_known_nodes()

        self.w_heard_peers = UpdatingText(self.app, "Heard Peers: ", get_num_peers, append_text=" (30m)")
        self.w_known_nodes = UpdatingText(self.app, "Known Nodes: ", get_num_nodes)

        pile = urwid.Pile([
            self.w_heard_peers,
            self.w_known_nodes,
        ])

        self.display_widget = urwid.LineBox(pile, title="Network Stats")

        super().__init__(self.display_widget)

    def start(self):
        self.w_heard_peers.start()
        self.w_known_nodes.start()

class NetworkLeftPile(urwid.Pile):
    def keypress(self, size, key):
        if key == "ctrl l":
            self.parent.toggle_list()
        elif key == "ctrl g":
            self.parent.toggle_fullscreen()
        elif key == "ctrl e":
            self.parent.selected_node_info()
        elif key == "ctrl p":
            self.parent.reinit_lxmf_peers()
            self.parent.show_peers()
        elif key == "ctrl w":
            self.parent.browser.disconnect()
        elif key == "ctrl u":
            self.parent.browser.url_dialog()
        elif key == "ctrl s":
            self.parent.browser.save_node_dialog()

        else:
            return super(NetworkLeftPile, self).keypress(size, key)


class NetworkDisplay():
    list_width = 0.33
    given_list_width = 52

    def __init__(self, app):
        self.app = app
        g = self.app.ui.glyphs

        self.browser = Browser(self.app, "nomadnetwork", "node", auth_identity = self.app.identity, delegate = self)

        if self.app.node != None:
            self.browser.loopback = self.app.node.destination.hash

        self.known_nodes_display = KnownNodes(self.app)
        self.lxmf_peers_display = LXMFPeers(self.app)
        self.network_stats_display = NetworkStats(self.app, self)
        self.announce_stream_display = AnnounceStream(self.app, self)
        self.local_peer_display = LocalPeer(self.app, self)
        self.node_info_display = NodeInfo(self.app, self)

        self.known_nodes_display.delegate = self

        self.list_display = 1
        self.left_pile = NetworkLeftPile([
            (urwid.WEIGHT, 1, self.known_nodes_display),
            # (urwid.PACK, self.network_stats_display),
            (urwid.PACK, self.local_peer_display),
        ])

        self.left_pile.parent = self

        self.left_area = self.left_pile
        self.right_area = self.browser.display_widget
        self.right_area_width = 1-NetworkDisplay.list_width

        self.columns = urwid.Columns(
            [
                # (urwid.WEIGHT, NetworkDisplay.list_width, self.left_area),
                # (urwid.WEIGHT, self.right_area_width, self.right_area)
                (NetworkDisplay.given_list_width, self.left_area),
                (urwid.WEIGHT, 1, self.right_area)
            ],
            dividechars=0, focus_column=0
        )

        self.shortcuts_display = NetworkDisplayShortcuts(self.app)
        self.widget = self.columns

    def toggle_list(self):
        if self.list_display != 0:
            options = self.left_pile.options(height_type=urwid.WEIGHT, height_amount=1)
            self.left_pile.contents[0] = (self.announce_stream_display, options)
            self.list_display = 0
        else:
            options = self.left_pile.options(height_type=urwid.WEIGHT, height_amount=1)
            self.left_pile.contents[0] = (self.known_nodes_display, options)
            self.list_display = 1

    def toggle_fullscreen(self):
        if NetworkDisplay.given_list_width != 0:
            self.saved_list_width = NetworkDisplay.given_list_width
            NetworkDisplay.given_list_width = 0
        else:
            NetworkDisplay.given_list_width = self.saved_list_width

        options = self.widget.options("given", NetworkDisplay.given_list_width)
        self.widget.contents[0] = (self.left_area, options)

    def show_peers(self):
        options = self.left_pile.options(height_type=urwid.WEIGHT, height_amount=1)
        self.left_pile.contents[0] = (self.lxmf_peers_display, options)

        if self.list_display != 0:
            self.list_display = 0
        else:
            self.list_display = 1

    def selected_node_info(self):
        if self.list_display == 1:
            parent = self.app.ui.main_display.sub_displays.network_display
            selected_node_entry = parent.known_nodes_display.ilb.get_selected_item()
            if selected_node_entry is not None:
                selected_node_hash = selected_node_entry.base_widget.display_widget.source_hash

                if selected_node_hash is not None:
                    info_widget = KnownNodeInfo(selected_node_hash)
                    options = parent.left_pile.options(height_type=urwid.WEIGHT, height_amount=1)
                    parent.left_pile.contents[0] = (info_widget, options)

    def focus_lists(self):
        self.columns.focus_position = 0

    def reinit_known_nodes(self):
        self.known_nodes_display = KnownNodes(self.app)
        self.known_nodes_display.delegate = self
        self.announce_stream_display.rebuild_widget_list()

    def reinit_lxmf_peers(self):
        if self.lxmf_peers_display:
            si = self.lxmf_peers_display.ilb.get_selected_position()
        else:
            si = None
        self.lxmf_peers_display = LXMFPeers(self.app)
        self.lxmf_peers_display.delegate = self
        self.close_list_dialogs()
        if si != None:
            self.lxmf_peers_display.ilb.select_item(si)

    def close_list_dialogs(self):
        if self.list_display == 0:
            options = self.left_pile.options(height_type=urwid.WEIGHT, height_amount=1)
            self.left_pile.contents[0] = (self.announce_stream_display, options)
        else:
            options = self.left_pile.options(height_type=urwid.WEIGHT, height_amount=1)
            self.left_pile.contents[0] = (self.known_nodes_display, options)

    def start(self):
        self.local_peer_display.start()
        self.node_info_display.start()
        self.network_stats_display.start()

    def shortcuts(self):
        return self.shortcuts_display

    def directory_change_callback(self):
        self.announce_stream_display.rebuild_widget_list()
        if self.known_nodes_display.no_content:
            self.reinit_known_nodes()
        else:
            self.known_nodes_display.rebuild_widget_list()


class LXMFPeers(urwid.WidgetWrap):
    def __init__(self, app):
        self.app = app
        self.peer_list = app.message_router.peers
        # self.peer_list = {}

        g = self.app.ui.glyphs

        self.widget_list = self.make_peer_widgets()

        self.ilb = IndicativeListBox(
            self.widget_list,
            on_selection_change=self.node_list_selection,
            initialization_is_selection_change=False,
            highlight_offFocus="list_off_focus"
        )

        if len(self.peer_list) > 0:
            self.display_widget = self.ilb
            widget_style = None
            self.no_content = False
        else:
            self.no_content = True
            widget_style = "inactive_text"
            self.pile = urwid.Pile([
                urwid.Text(("warning_text", g["info"]+"\n"), align=urwid.CENTER),
                SelectText(("warning_text", "Currently, no LXMF nodes are peered\n\n"), align=urwid.CENTER),
            ])
            self.display_widget = urwid.Filler(self.pile, valign=urwid.TOP, height=urwid.PACK)

        if hasattr(self, "peer_list") and self.peer_list:
            pl = len(self.peer_list)
        else:
            pl = 0
        super().__init__(urwid.AttrMap(urwid.LineBox(self.display_widget, title=f"LXMF Propagation Peers ({pl})"), widget_style))

    def keypress(self, size, key):
        if key == "up" and (self.no_content or self.ilb.first_item_is_selected()):
            nomadnet.NomadNetworkApp.get_shared_instance().ui.main_display.frame.focus_position = "header"
        elif key == "ctrl x":
            self.delete_selected_entry()
        elif key == "ctrl r":
            self.sync_selected_entry()

        return super(LXMFPeers, self).keypress(size, key)


    def node_list_selection(self, arg1, arg2):
        pass

    def delete_selected_entry(self):
        si = self.ilb.get_selected_item()
        if si != None:
            destination_hash = si.original_widget.destination_hash
            self.app.message_router.unpeer(destination_hash)
            self.delegate.reinit_lxmf_peers()
            self.delegate.show_peers()

    def sync_selected_entry(self):
        sync_grace = 10
        si = self.ilb.get_selected_item()
        if si != None:
            destination_hash = si.original_widget.destination_hash
            if destination_hash in self.app.message_router.peers:
                peer = self.app.message_router.peers[destination_hash]
                if time.time() > peer.last_sync_attempt+sync_grace:
                    peer.next_sync_attempt = time.time()-1

                    def job():
                        peer.sync()
                    threading.Thread(target=job, daemon=True).start()

                    time.sleep(0.25)

                    def dismiss_dialog(sender):
                        self.close_list_dialogs()

                    dialog = ListDialogLineBox(
                        urwid.Pile([
                            urwid.Text("A delivery sync of all unhandled LXMs was manually requested for the selected node\n", align=urwid.CENTER),
                            urwid.Columns([
                                (urwid.WEIGHT, 0.1, urwid.Text("")),
                                (urwid.WEIGHT, 0.45, urwid.Button("OK", on_press=dismiss_dialog)),
                            ])
                        ]),
                        title="!",

                    )
                    dialog.delegate = self.delegate
                    bottom = self

                    overlay = urwid.Overlay(dialog, bottom, align=urwid.CENTER, width=urwid.RELATIVE_100, valign=urwid.MIDDLE, height=urwid.PACK, left=2, right=2)

                    options = self.delegate.left_pile.options(urwid.WEIGHT, 1)
                    self.delegate.left_pile.contents[0] = (overlay, options)


    def close_list_dialogs(self):
        self.delegate.reinit_lxmf_peers()
        self.delegate.show_peers()

    def rebuild_widget_list(self):
        self.peer_list = self.app.message_router.peers
        self.widget_list = self.make_peer_widgets()
        self.ilb.set_body(self.widget_list)
        if len(self.widget_list) > 0:
            self.no_content = False
        else:
            self.no_content = True
            self.delegate.reinit_lxmf_peers()

    def make_peer_widgets(self):
        widget_list = []
        sorted_peers = sorted(self.peer_list, key=lambda pid: (self.app.directory.pn_trust_level(pid), self.peer_list[pid].sync_transfer_rate), reverse=True)
        for peer_id in sorted_peers:
            peer = self.peer_list[peer_id]
            trust_level = self.app.directory.pn_trust_level(peer_id)
            pe = LXMFPeerEntry(self.app, peer, self, trust_level)
            pe.destination_hash = peer.destination_hash
            widget_list.append(pe)

        return widget_list

class LXMFPeerEntry(urwid.WidgetWrap):
    def __init__(self, app, peer, delegate, trust_level):
        destination_hash = peer.destination_hash

        self.app = app
        g = self.app.ui.glyphs

        node_identity = RNS.Identity.recall(destination_hash)
        display_str = RNS.prettyhexrep(destination_hash)
        if node_identity != None:
            node_hash = RNS.Destination.hash_from_name_and_identity("nomadnetwork.node", node_identity)
            display_name = self.app.directory.alleged_display_str(node_hash)
            if display_name != None:
                display_str = str(display_name)+"\n  "+display_str

        sym = g["sent"]
        style         = "list_unknown"
        focus_style   = "list_focus"

        alive_string = "Unknown"
        if hasattr(peer, "alive"):
            if peer.alive:
                alive_string = "Available"
                if trust_level == DirectoryEntry.TRUSTED:
                    style = "list_trusted"
                    focus_style = "list_focus_trusted"
                else:
                    style = "list_normal"
                    focus_style = "list_focus"
            else:
                alive_string = "Unresponsive"
                style = "list_unresponsive"
                focus_style = "list_focus_unresponsive"

        if peer.propagation_transfer_limit: txfer_limit = RNS.prettysize(peer.propagation_transfer_limit*1000)
        else:                               txfer_limit = "No"
        
        if peer.propagation_sync_limit: sync_limit = RNS.prettysize(peer.propagation_sync_limit*1000)
        else:                           sync_limit = "Unknown"
        
        if peer.propagation_stamp_cost: sct = peer.propagation_stamp_cost
        else:                           sct = "Unknown"
        
        if peer.propagation_stamp_cost_flexibility: scf = f" (flex {peer.propagation_stamp_cost_flexibility})"
        else:                                       scf = ""
        
        ar = round(peer.acceptance_rate*100, 2)
        peer_info_str  = sym+" "+display_str+"\n  "+alive_string+", last heard "+pretty_date(int(peer.last_heard))
        peer_info_str += f"\n  {sync_limit} sync limit, {txfer_limit} msg limit"
        peer_info_str += f"\n  {RNS.prettyspeed(peer.sync_transfer_rate)} STR, {RNS.prettyspeed(peer.link_establishment_rate)} LER"
        peer_info_str += f"\n  Propagation cost {sct}{scf}"
        peer_info_str += "\n  "+str(peer.unhandled_message_count)+f" unhandled LXMs, {ar}% AR"
        widget = ListEntry(peer_info_str)
        self.display_widget = urwid.AttrMap(widget, style, focus_style)
        self.display_widget.destination_hash = destination_hash
        super().__init__(self.display_widget)


def pretty_date(time=False):
    """
    Get a datetime object or a int() Epoch timestamp and return a
    pretty string like 'an hour ago', 'Yesterday', '3 months ago',
    'just now', etc
    """
    from datetime import datetime
    now = datetime.now()
    if type(time) is int:
        diff = now - datetime.fromtimestamp(time)
    elif isinstance(time,datetime):
        diff = now - time
    elif not time:
        diff = now - now
    second_diff = diff.seconds
    day_diff = diff.days

    if day_diff < 0:
        return ''

    if day_diff == 0:
        if second_diff < 10:
            return "just now"
        if second_diff < 60:
            return str(second_diff) + " seconds ago"
        if second_diff < 120:
            return "a minute ago"
        if second_diff < 3600:
            return str(int(second_diff / 60)) + " minutes ago"
        if second_diff < 7200:
            return "an hour ago"
        if second_diff < 86400:
            return str(int(second_diff / 3600)) + " hours ago"
    if day_diff == 1:
        return "Yesterday"
    if day_diff < 7:
        return str(day_diff) + " days ago"
    if day_diff < 31:
        return str(int(day_diff / 7)) + " weeks ago"
    if day_diff < 365:
        return str(int(day_diff / 30)) + " months ago"
    return str(int(day_diff / 365)) + " years ago"


import RNS

from .Network import *
from .Conversations import *
from .Channels import *
from .Directory import *
from .Config import *
from .Interfaces import *
from .Map import *
from .Log import *
from .Guide import *
import urwid

class SubDisplays():
    def __init__(self, app):
        self.app = app
        self.network_display = NetworkDisplay(self.app)
        self.conversations_display = ConversationsDisplay(self.app)
        self.channels_display = ChannelsDisplay(self.app)
        self.directory_display = DirectoryDisplay(self.app)
        self.config_display = ConfigDisplay(self.app)
        self.interface_display = InterfaceDisplay(self.app)
        self.map_display = MapDisplay(self.app)
        self.log_display = LogDisplay(self.app)
        self.guide_display = GuideDisplay(self.app)

        if app.firstrun:
            self.active_display = self.guide_display
        else:
            self.active_display = self.conversations_display

    def active(self):
        return self.active_display

class MenuButton(urwid.Button):
    button_left = urwid.Text('[')
    button_right = urwid.Text(']')

class MainFrame(urwid.Frame):
    FOCUS_CHECK_TIMEOUT = 0.25

    def __init__(self, body, header=None, footer=None, delegate=None):
        self.delegate = delegate
        self.current_focus = None
        super().__init__(body, header, footer)

    def keypress_focus_check(self, deferred=False):
        current_focus = self.delegate.widget.get_focus_widgets()[-1]

        if deferred:
            if current_focus != self.current_focus:
                self.focus_changed()
        else:
            def deferred_focus_check(loop, user_data):
                self.keypress_focus_check(deferred=True)
            self.delegate.app.ui.loop.set_alarm_in(MainFrame.FOCUS_CHECK_TIMEOUT, deferred_focus_check)

        self.current_focus = current_focus

    def focus_changed(self):
        current_focus = self.delegate.widget.get_focus_widgets()[-1]
        current_focus_path = self.delegate.widget.get_focus_path()
        
        if len(current_focus_path) > 1:
            if current_focus_path[0] == "body":
                self.delegate.update_active_shortcuts()

        if self.delegate.sub_displays.active() == self.delegate.sub_displays.conversations_display:
            # Needed to refresh indicativelistbox styles on mouse focus change
            self.delegate.sub_displays.conversations_display.focus_change_event()

    def mouse_event(self, size, event, button, col, row, focus):
        current_focus = self.delegate.widget.get_focus_widgets()[-1]
        if current_focus != self.current_focus:
            self.focus_changed()

        self.current_focus = current_focus
        return super(MainFrame, self).mouse_event(size, event, button, col, row, focus)

    def keypress(self, size, key):
        self.keypress_focus_check()
        
        #if key == "ctrl q":
        #    raise urwid.ExitMainLoop

        return super(MainFrame, self).keypress(size, key)

class MainDisplay():
    def __init__(self, ui, app):
        self.ui = ui
        self.app = app

        self.menu_display = MenuDisplay(self.app, self)
        self.sub_displays = SubDisplays(self.app)

        self.frame = MainFrame(self.sub_displays.active().widget, header=self.menu_display.widget, footer=self.sub_displays.active().shortcuts().widget, delegate=self)
        self.widget = self.frame

    def show_network(self, user_data):
        self.sub_displays.active_display = self.sub_displays.network_display
        self.update_active_sub_display()
        self.sub_displays.network_display.start()

    def show_conversations(self, user_data):
        self.sub_displays.active_display = self.sub_displays.conversations_display
        self.update_active_sub_display()

    def show_channels(self, user_data):
        self.sub_displays.active_display = self.sub_displays.channels_display
        self.update_active_sub_display()
        self.sub_displays.channels_display.start()

    def show_directory(self, user_data):
        self.sub_displays.active_display = self.sub_displays.directory_display
        self.update_active_sub_display()

    def show_map(self, user_data):
        self.sub_displays.active_display = self.sub_displays.map_display
        self.update_active_sub_display()

    def show_config(self, user_data):
        self.sub_displays.active_display = self.sub_displays.config_display
        self.update_active_sub_display()

    def show_interfaces(self, user_data):
        self.sub_displays.active_display = self.sub_displays.interface_display
        self.update_active_sub_display()
        self.sub_displays.interface_display.start()

    def show_log(self, user_data):
        self.sub_displays.active_display = self.sub_displays.log_display
        self.sub_displays.log_display.show()
        self.update_active_sub_display()

    def show_guide(self, user_data):
        self.sub_displays.active_display = self.sub_displays.guide_display
        self.update_active_sub_display()

    def update_active_sub_display(self):
        self.frame.contents["body"] = (self.sub_displays.active().widget, None)
        self.update_active_shortcuts()
        if self.sub_displays.active_display != self.sub_displays.log_display:
            self.sub_displays.log_display.kill()

    def update_active_shortcuts(self):
        self.frame.contents["footer"] = (self.sub_displays.active().shortcuts().widget, None)

    def request_redraw(self, extra_delay=0.0):
        self.app.ui.loop.set_alarm_in(0.25+extra_delay, self.redraw_now)
    
    def redraw_now(self, sender=None, data=None):
        self.app.ui.loop.screen.clear()
        #self.app.ui.loop.draw_screen()

    def start(self):
        self.menu_display.start()

    def quit(self, sender=None):
        logterm_pid = None
        if True or RNS.vendor.platformutils.is_android():
            if self.sub_displays.log_display != None and self.sub_displays.log_display.log_term != None:
                if self.sub_displays.log_display.log_term.log_term != None:
                    logterm_pid = self.sub_displays.log_display.log_term.log_term.pid
                    if logterm_pid != None:
                        import os, signal
                        os.kill(logterm_pid, signal.SIGKILL)

        raise urwid.ExitMainLoop


class MenuColumns(urwid.Columns):
    def keypress(self, size, key):
        if key == "tab" or key == "down":
            self.handler.frame.focus_position = "body"

        return super(MenuColumns, self).keypress(size, key)

class MenuDisplay():
    UPDATE_INTERVAL = 2

    def __init__(self, app, handler):
        self.app = app
        self.update_interval = MenuDisplay.UPDATE_INTERVAL
        self.g = self.app.ui.glyphs

        self.menu_indicator  = urwid.Text("")

        menu_text             = (urwid.PACK, self.menu_indicator)
        button_network        = (11, MenuButton("Network", on_press=handler.show_network))
        button_conversations  = (17, MenuButton("Conversations", on_press=handler.show_conversations))
        button_channels       = (12, MenuButton("Channels", on_press=handler.show_channels))
        button_directory      = (13, MenuButton("Directory", on_press=handler.show_directory))
        button_map            = (7,  MenuButton("Map", on_press=handler.show_map))
        button_log            = (7,  MenuButton("Log", on_press=handler.show_log))
        button_config         = (10, MenuButton("Config", on_press=handler.show_config))
        button_interfaces     = (14, MenuButton("Interfaces", on_press=handler.show_interfaces))
        button_guide          = (9,  MenuButton("Guide", on_press=handler.show_guide))
        button_quit           = (8,  MenuButton("Quit", on_press=handler.quit))

        # buttons = [menu_text, button_conversations, button_node, button_directory, button_map]
        if self.app.config["textui"]["hide_guide"]:
            buttons = [menu_text, button_conversations, button_network, button_channels, button_log, button_interfaces, button_config, button_quit]
        else:
            buttons = [menu_text, button_conversations, button_network, button_channels, button_log, button_interfaces, button_config, button_guide, button_quit]

        columns = MenuColumns(buttons, dividechars=1)
        columns.handler = handler

        self.update_display()

        self.widget = urwid.AttrMap(columns, "menubar")

    def start(self):
        self.update_display_job()

    def update_display_job(self, event = None, sender = None):
        self.update_display()
        self.app.ui.loop.set_alarm_in(self.update_interval, self.update_display_job)

    def update_display(self):
        if self.app.has_unread_conversations():
            self.indicate_unread()
        else:
            self.indicate_normal()

    def indicate_normal(self):
        self.menu_indicator.set_text(self.g["decoration_menu"])

    def indicate_unread(self):
        self.menu_indicator.set_text(self.g["unread_menu"])



import sys
import base64
import urwid
import RNS


def osc52_copy(text):
    if not text:
        return False
    try:
        encoded = base64.b64encode(text.encode("utf-8")).decode("ascii")
        sys.stdout.write("\x1b]52;c;" + encoded + "\x07")
        sys.stdout.flush()
        return True
    except Exception as e:
        RNS.log("Could not emit clipboard escape sequence: "+str(e), RNS.LOG_ERROR)
        return False


class ClickableIcon(urwid.Text):
    _selectable = False

    def __init__(self, text, on_click=None):
        super().__init__(text)
        self._on_click = on_click

    def mouse_event(self, size, event, button, x, y, focus):
        if button == 1 and "press" in event and self._on_click is not None:
            self._on_click()
            return True
        return False
